#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Automated scanning module functionalities for PyMoDAQ

Contains all objects related to the DAQScan module, to do automated scans, saving data...
"""
from __future__ import annotations
from collections import OrderedDict
import logging
import os
from pathlib import Path
import sys
import tempfile
from typing import List, Tuple, TYPE_CHECKING

import numpy as np
from qtpy import QtWidgets, QtCore, QtGui
from qtpy.QtCore import QObject, Slot, QThread, Signal, QDateTime, QDate, QTime

from pymodaq.utils import data as data_mod
from pymodaq.utils.logger import set_logger, get_module_name
from pymodaq.utils.config import Config, get_set_preset_path
from pymodaq.utils.parameter import ioxml
from pymodaq.utils.plotting.data_viewers.viewer import ViewersEnum
from pymodaq.utils.managers.parameter_manager import ParameterManager, Parameter, ParameterTree

from pymodaq.utils import exceptions
from pymodaq.utils.plotting.data_viewers.viewer2D import Viewer2D
from pymodaq.utils.plotting.data_viewers.viewer1D import Viewer1D
from pymodaq.utils.plotting.navigator import Navigator
from pymodaq.utils.plotting.scan_selector import ScanSelector, SelectorItem
from pymodaq.utils.scanner.scanner import Scanner, scanner_factory  #, adaptive, adaptive_losses
from pymodaq.utils.managers.batchscan_manager import BatchScanner
from pymodaq.utils.managers.modules_manager import ModulesManager
from pymodaq.post_treatment.load_and_plot import LoaderPlotter
from pymodaq.utils.messenger import messagebox
from pymodaq.extensions.daq_scan_ui import DAQScanUI

from pymodaq.utils import daq_utils as utils
from pymodaq.utils import gui_utils as gutils
from pymodaq.utils.h5modules.saving import H5Saver
from pymodaq.utils.h5modules import module_saving, data_saving
from pymodaq.utils.data import DataToExport, DataActuator

if TYPE_CHECKING:
    from pymodaq.dashboard import DashBoard

config = Config()
logger = set_logger(get_module_name(__file__))

SHOW_POPUPS = config('scan', 'show_popups')


class ScanDataTemp:
    """Convenience class to hold temporary data to be plotted in the live plots"""
    def __init__(self, scan_index: int, indexes: Tuple[int], data: data_mod.DataToExport):
        self.scan_index = scan_index
        self.indexes = indexes
        self.data = data


class DAQScan(QObject, ParameterManager):
    """
    Main class initializing a DAQScan module with its dashboard and scanning control panel
    """
    settings_name = 'daq_scan_settings'
    command_daq_signal = Signal(utils.ThreadCommand)
    status_signal = Signal(str)
    live_data_1D_signal = Signal(list)

    params = [
        {'title': 'Time Flow:', 'name': 'time_flow', 'type': 'group', 'expanded': False,
         'children': [
            {'title': 'Wait time step (ms)', 'name': 'wait_time', 'type': 'int', 'value': 0,
             'tip': 'Wait time in ms after each step of acquisition (move and grab)'},
            {'title': 'Wait time between (ms)', 'name': 'wait_time_between', 'type': 'int',
             'value': 0,
             'tip': 'Wait time in ms between move and grab processes'},
            {'title': 'Timeout (ms)', 'name': 'timeout', 'type': 'int', 'value': 10000},
        ]},
        {'title': 'Scan options', 'name': 'scan_options', 'type': 'group', 'children': [
            {'title': 'Naverage:', 'name': 'scan_average', 'type': 'int', 'value': 1, 'min': 1},
        ]},

        {'title': 'Plotting options', 'name': 'plot_options', 'type': 'group', 'children': [
            {'title': 'Get data', 'name': 'plot_probe', 'type': 'bool_push'},
            {'title': 'Group 0D data:', 'name': 'group0D', 'type': 'bool', 'value': True},
            {'title': 'Plot 0Ds:', 'name': 'plot_0d', 'type': 'itemselect', 'checkbox': True},
            {'title': 'Plot 1Ds:', 'name': 'plot_1d', 'type': 'itemselect', 'checkbox': True},
            {'title': 'Prepare Viewers', 'name': 'prepare_viewers', 'type': 'bool_push'},
            {'title': 'Plot at each step?', 'name': 'plot_at_each_step', 'type': 'bool',
             'value': True},
            {'title': 'Refresh Plots (ms)', 'name': 'refresh_live', 'type': 'int',
             'value': 1000, 'visible': False},
            ]},
    ]

    def __init__(self, dockarea: gutils.DockArea = None, dashboard: DashBoard = None):
        """

        Parameters
        ----------
        dockarea: DockArea
            instance of the modified pyqtgraph Dockarea
        dashboard: DashBoard
            instance of the pymodaq dashboard

        """
        
        logger.info('Initializing DAQScan')
        QObject.__init__(self)
        ParameterManager.__init__(self)

        self.title = __class__.__name__

        self.dockarea: gutils.DockArea = dockarea
        self.dashboard: DashBoard = dashboard
        if dashboard is None:
            raise Exception('No valid dashboard initialized')

        self.mainwindow = self.dockarea.parent()
        self.ui: DAQScanUI = DAQScanUI(self.dockarea)

        self.wait_time = 1000

        self.navigator: Navigator = None
        self.scan_selector: ScanSelector = None

        self.ind_scan = 0
        self.ind_average = 0

        self._metada_dataset_set = False

        self.curvilinear_values = []
        self.plot_colors = utils.plot_colors

        self.scan_thread: QThread = None

        self.modules_manager = ModulesManager(self.dashboard.detector_modules, self.dashboard.actuators_modules)
        self.modules_manager.settings.child('data_dimensions').setOpts(expanded=False)
        self.modules_manager.settings.child('actuators_positions').setOpts(expanded=False)
        self.modules_manager.detectors_changed.connect(self.clear_plot_from)

        self.h5saver = H5Saver()
        self.module_and_data_saver = module_saving.ScanSaver(self)
        self.module_and_data_saver.h5saver = self.h5saver

        self.extended_saver: data_saving.DataToExportExtendedSaver = None
        self.h5temp: H5Saver = None
        self.temp_path: tempfile.TemporaryDirectory = None

        self.h5saver.settings.child('do_save').hide()
        self.h5saver.settings.child('custom_name').hide()
        self.h5saver.new_file_sig.connect(self.create_new_file)

        self.scanner = Scanner(actuators=self.modules_manager.actuators)  # , adaptive_losses=adaptive_losses)
        self.scan_parameters = None

        self.batcher: BatchScanner = None
        self.batch_started = False
        self.ind_batch = 0

        self.modules_manager.actuators_changed[list].connect(self.update_actuators)

        self.setup_ui()
        self.ui.command_sig.connect(self.process_ui_cmds)

        self.create_dataset_settings()

        self.set_config()

        self.live_plotter = LoaderPlotter(self.dockarea)
        self.live_timer = QtCore.QTimer()
        self.live_timer.timeout.connect(self.update_live_plots)

        self.ui.enable_start_stop(True)
        logger.info('DAQScan Initialized')

    def plot_from(self):
        self.modules_manager.get_det_data_list()
        data0D = self.modules_manager.settings['data_dimensions', 'det_data_list0D']
        data1D = self.modules_manager.settings['data_dimensions', 'det_data_list1D']
        data0D['selected'] = data0D['all_items']
        data1D['selected'] = data1D['all_items']
        self.settings.child('plot_options', 'plot_0d').setValue(data0D)
        self.settings.child('plot_options', 'plot_1d').setValue(data1D)

    def setup_ui(self):
        self.ui.populate_toolbox_widget([self.settings_tree, self.h5saver.settings_tree],
                                        ['General Settings', 'Save Settings'])

        self.ui.set_scanner_settings(self.scanner.parent_widget)
        self.ui.set_modules_settings(self.modules_manager.settings_tree)

        self.plotting_settings_tree = ParameterTree()
        self.plotting_settings_tree.setParameters(self.settings.child('plot_options'))
        self.ui.set_plotting_settings(self.plotting_settings_tree)

    ################
    #  CONFIG/SETUP UI / EXIT

    def set_config(self):
        self.settings.child('time_flow', 'wait_time').setValue(config['scan']['timeflow']['wait_time'])
        self.settings.child('time_flow', 'wait_time_between').setValue(config['scan']['timeflow']['wait_time'])
        self.settings.child('time_flow', 'timeout').setValue(config['scan']['timeflow']['timeout'])

        self.settings.child('scan_options',  'scan_average').setValue(config['scan']['Naverage'])

    def process_ui_cmds(self, cmd: utils.ThreadCommand):
        """Process commands sent by actions done in the ui

        Parameters
        ----------
        cmd: ThreadCommand
            Possible values are:
                * quit
                * ini_positions
                * start
                * start_batch
                * stop
                * move_at
                * show_log
                * load
                * save
                * show_file
                * navigator
                * batch
                * viewers_changed
        """
        if cmd.command == 'quit':
            self.quit_fun()
        elif cmd.command == 'ini_positions':
            self.set_ini_positions()
        elif cmd.command == 'start':
            self.start_scan()
        elif cmd.command == 'start_batch':
            self.start_scan_batch()
        elif cmd.command == 'stop':
            self.stop_scan()
        elif cmd.command == 'move_at':
            self.move_to_crosshair()
        elif cmd.command == 'show_log':
            self.show_log()
        elif cmd.command == 'load':
            self.load_file()
        elif cmd.command == 'save':
            self.save_file()
        elif cmd.command == 'show_file':
            self.show_file_content()
        elif cmd.command == 'navigator':
            self.show_navigator()
        elif cmd.command == 'batch':
            self.show_batcher(self.ui.menubar)
        elif cmd.command == 'viewers_changed':
            ...

    def show_log(self):
        """Open the log file in the default text editor"""
        import webbrowser
        webbrowser.open(logger.parent.handlers[0].baseFilename)

    def quit_fun(self):
        """
            Quit the current instance of DAQ_scan

            See Also
            --------
            quit_fun
        """
        try:
            if self.temp_path is not None:
                try:
                    self.h5temp.close()
                    self.temp_path.cleanup()
                except Exception as e:
                    logger.exception(str(e))

            self.h5saver.close_file()
            self.mainwindow.close()

        except Exception as e:
            logger.exception(str(e))

    def create_dataset_settings(self):
        # params about dataset attributes and scan attibutes
        date = QDateTime(QDate.currentDate(), QTime.currentTime())
        params_dataset = [{'title': 'Dataset information', 'name': 'dataset_info', 'type': 'group', 'children': [
            {'title': 'Author:', 'name': 'author', 'type': 'str', 'value': config['user']['name']},
            {'title': 'Date/time:', 'name': 'date_time', 'type': 'date_time', 'value': date},
            {'title': 'Sample:', 'name': 'sample', 'type': 'str', 'value': ''},
            {'title': 'Experiment type:', 'name': 'experiment_type', 'type': 'str', 'value': ''},
            {'title': 'Description:', 'name': 'description', 'type': 'text', 'value': ''}]}]

        params_scan = [{'title': 'Scan information', 'name': 'scan_info', 'type': 'group', 'children': [
            {'title': 'Author:', 'name': 'author', 'type': 'str', 'value': config['user']['name']},
            {'title': 'Date/time:', 'name': 'date_time', 'type': 'date_time', 'value': date},
            {'title': 'Scan type:', 'name': 'scan_type', 'type': 'str', 'value': ''},
            {'title': 'Scan subtype:', 'name': 'scan_sub_type', 'type': 'str', 'value': ''},
            {'title': 'Scan name:', 'name': 'scan_name', 'type': 'str', 'value': '', 'readonly': True},
            {'title': 'Description:', 'name': 'description', 'type': 'text', 'value': ''},
        ]}]

        self.dataset_attributes = Parameter.create(name='Attributes', type='group', children=params_dataset)
        self.scan_attributes = Parameter.create(name='Attributes', type='group', children=params_scan)
    ###################
    # external modules

    def show_batcher(self, menubar):
        self.batcher = BatchScanner(self.dockarea, self.modules_manager.actuators_all,
                                    self.modules_manager.detectors_all)
        self.batcher.create_menu(menubar)
        self.batcher.setupUI()
        self.ui.set_action_visible('start_batch', True)

    def start_scan_batch(self):
        self.batch_started = True
        self.ind_batch = 0
        self.loop_scan_batch()

    def loop_scan_batch(self):
        if self.ind_batch >= len(self.batcher.scans_names):
            self.stop_scan()
            return
        self.scanner = self.batcher.get_scan(self.batcher.scans_names[self.ind_batch])
        actuators, detectors = self.batcher.get_act_dets()
        self.set_scan_batch(actuators[self.batcher.scans_names[self.ind_batch]],
                            detectors[self.batcher.scans_names[self.ind_batch]])
        self.start_scan()

    def set_scan_batch(self, actuators, detectors):
        self.modules_manager.selected_detectors_name = detectors
        self.modules_manager.selected_actuators_name = actuators
        QtWidgets.QApplication.processEvents()

    def show_file_attributes(self, type_info='dataset'):
        """
            Switch the type_info value.

            In case of :
                * *scan* : Set parameters showing top false
                * *dataset* : Set parameters showing top false
                * *managers* : Set parameters showing top false. Add the save/cancel buttons to the accept/reject dialog (to save managers parameters in a xml file).

            Finally, in case of accepted managers type info, save the managers parameters in a xml file.

            =============== =========== ====================================
            **Parameters**    **Type**    **Description**
            *type_info*       string      The file type information between
                                            * scan
                                            * dataset
                                            * managers
            =============== =========== ====================================

            See Also
            --------
            custom_tree.parameter_to_xml_file, create_menu
        """
        if SHOW_POPUPS:
            dialog = QtWidgets.QDialog()
            vlayout = QtWidgets.QVBoxLayout()
            tree = ParameterTree()
            tree.setMinimumWidth(400)
            tree.setMinimumHeight(500)
            if type_info == 'scan':
                tree.setParameters(self.scan_attributes, showTop=False)
            elif type_info == 'dataset':
                tree.setParameters(self.dataset_attributes, showTop=False)

            vlayout.addWidget(tree)
            dialog.setLayout(vlayout)
            buttonBox = QtWidgets.QDialogButtonBox(parent=dialog)
            buttonBox.addButton('Cancel', buttonBox.RejectRole)
            buttonBox.addButton('Apply', buttonBox.AcceptRole)
            buttonBox.rejected.connect(dialog.reject)
            buttonBox.accepted.connect(dialog.accept)

            vlayout.addWidget(buttonBox)
            dialog.setWindowTitle('Fill in information about this {}'.format(type_info))
            res = dialog.exec()
        else:
            res = True
        return res

    def show_file_content(self):
        try:
            self.h5saver.init_file(addhoc_file_path=self.h5saver.settings.child(('current_h5_file')).value())
            self.h5saver.show_file_content()
        except Exception as e:
            logger.exception(str(e))

    def show_navigator(self):

        if self.navigator is None:
            # loading navigator
            self.navigator_dock = gutils.Dock('Navigator')
            widgnav = QtWidgets.QWidget()
            self.navigator_dock.addWidget(widgnav)
            self.dockarea.addDock(self.navigator_dock)
            self.navigator_dock.float()

            self.navigator = Navigator(widgnav)

            self.navigator.log_signal[str].connect(self.dashboard.add_status)
            self.navigator.settings.child('settings', 'Load h5').hide()
            self.navigator.set_action_visible('load_scan', False)

            self.navigator.sig_double_clicked.connect(self.move_at)
            self.navigator.h5saver = self.h5saver
            self.navigator.list_2D_scans()

        self.show_scan_selector()

    def show_scan_selector(self):
        viewer_items = []
        if self.navigator is not None:
            viewer_items.append(SelectorItem(self.navigator.viewer, name='Navigator'))
        #
        # for viewer in self.live_plotter.viewers:
        #     viewer_items.update({viewer.title: dict(viewers=[viewer], names=[viewer.title])})
        self.scan_selector = ScanSelector(viewer_items)

        self.ui.add_scanner_settings(self.scan_selector.settings_tree)

        self.scan_selector.scan_select_signal.connect(self.scanner.update_from_scan_selector)

    ################
    #  LOADING SAVING

    def load_file(self):
        self.h5saver.load_file(self.h5saver.h5_file_path)

    def save_file(self):
        if not os.path.isdir(self.h5saver.settings['base_path']):
            os.mkdir(self.h5saver.settings['base_path'])
        filename = gutils.file_io.select_file(self.h5saver.settings['base_path'], save=True, ext='h5')
        self.h5saver.h5_file.copy_file(str(filename))

    def save_metadata(self, node, type_info='dataset_info'):
        """
            Switch the type_info value with :
                * *'dataset_info'* : Give the params attributes the dataset_attributes values
                * *'dataset'* : Give the params attributes the scan_attributes values

            |
            | Once done, course the params and add string casted date/time metadata as an element of attributes array.
            | Save the contents of given parameter object into a xml string unde the attributes settings.

            =============== =================== =========================================
            **Parameters**    **Type**           **Description**
            *node*            pytables h5 node   Root node to be treated
            *type_info*       string             File type info between :
                                                    * 'dataset_info'
                                                    * 'scan_info'
            =============== =================== =========================================

            See Also
            --------
            custom_tree.parameter_to_xml_string
        """

        attr = node.attrs
        if type_info == 'dataset_info':
            attr['type'] = 'dataset'
            params = self.dataset_attributes
        else:
            attr['type'] = 'scan'
            params = self.scan_attributes
        for child in params.child(type_info).children():
            if type(child.value()) is QDateTime:
                attr[child.name()] = child.value().toString('dd/mm/yyyy HH:MM:ss')
            else:
                attr[child.name()] = child.value()
        if type_info == 'dataset_info':
            # save contents of given parameter object into an xml string under the attribute settings
            settings_str = b'<All_settings title="All Settings" type="group">' + \
                           ioxml.parameter_to_xml_string(params) + \
                           ioxml.parameter_to_xml_string(self.settings)
                           # ioxml.parameter_to_xml_string(
                           #     self.dashboard.preset_manager.preset_params) +\
            settings_str += b'</All_settings>'
            attr['settings'] = settings_str

        elif type_info == 'scan_info':
            settings_all = [ioxml.parameter_to_xml_string(params),
                           ioxml.parameter_to_xml_string(self.settings),
                           ioxml.parameter_to_xml_string(self.h5saver.settings),
                           ioxml.parameter_to_xml_string(self.scanner.settings)]

            settings_str = b'<All_settings title="All Settings" type="group">'
            for set in settings_all:
                if len(settings_str + set) < 60000:
                    # size limit for any object header (including all the other attributes) is 64kb
                    settings_str += set
                else:
                    break
            settings_str += b'</All_settings>'
            attr['settings'] = settings_str

    def create_new_file(self, new_file):
        if new_file:
            self._metada_dataset_set = False
            self.h5saver.close_file()
        self.h5saver.init_file(update_h5=new_file)
        self.module_and_data_saver.h5saver = self.h5saver
        res = self.update_file_settings()
        if new_file:
            self.ui.enable_start_stop()
        return res

    def update_file_settings(self):
        try:
            res = True
            if not self._metada_dataset_set:
                res = self.set_metadata_about_dataset()
                self.save_metadata(self.h5saver.raw_group, 'dataset_info')

            if self.navigator is not None:
                self.navigator.update_h5file(self.h5saver.h5_file)
                self.navigator.settings.child('settings', 'filepath').setValue(self.h5saver.h5_file.filename)

            # # set attributes to the current group, such as scan_type....
            # self.scan_attributes.child('scan_info', 'scan_type').setValue(
            #     self.scanner.settings.child('scan_type').value())
            # self.scan_attributes.child('scan_info', 'scan_sub_type').setValue(
            #     self.scanner.settings.child('scan_sub_type').value())
            #
            # scan_node = self.module_and_data_saver.get_set_node()
            # self.scan_attributes.child('scan_info', 'scan_name').setValue(scan_node.name)
            # self.scan_attributes.child('scan_info', 'description').setValue('')
            #
            # res = self.set_metadata_about_current_scan()

            return res

        except Exception as e:
            logger.exception(str(e))

    def update_scan_info(self):
        # set attributes to the current group, such as scan_type....
        self.scan_attributes.child('scan_info', 'scan_type').setValue(
            self.scanner.settings.child('scan_type').value())
        self.scan_attributes.child('scan_info', 'scan_sub_type').setValue(
            self.scanner.settings.child('scan_sub_type').value())
        scan_node = self.module_and_data_saver.get_set_node(new=False)
        if scan_node.attrs['scan_done']:
            scan_name = self.module_and_data_saver.get_next_node_name()
        else:
            scan_name = scan_node.name
        self.scan_attributes.child('scan_info', 'scan_name').setValue(scan_name)
        self.scan_attributes.child('scan_info', 'description').setValue('')

        res = self.set_metadata_about_current_scan()
        return res

    #  PROCESS MODIFICATIONS
    def update_actuators(self, actuators: List[str]):
        self.scanner.actuators = self.modules_manager.actuators

    def move_to_crosshair(self, *args, **kwargs):
        if self.ui.is_action_checked('move_at'):
            self.modules_manager.connect_actuators()
            self.live_plotter.connect_double_clicked(self.move_at)
        else:
            self.live_plotter.disconnect(self.move_at)
            self.modules_manager.connect_actuators(False)

    def move_at(self, posx: float, posy: float = None):
        if logging.getLevelName(logger.level) == 'DEBUG':
            print(f'clicked at: {posx}, {posy}')
        positions = [posx, posy]
        positions = positions[:self.scanner.n_axes]
        self.modules_manager.move_actuators(positions)

    def value_changed(self, param):
        """

        """
        if param.name() == 'scan_average':
            self.ui.show_average_step(param.value() > 1)
        elif param.name() == 'prepare_viewers':
            self.prepare_viewers()
        elif param.name() == 'plot_probe':
            self.plot_from()
        elif param.name() == 'plot_at_each_step':
            self.settings.child('plot_options', 'refresh_live').show(not param.value())

    def clear_plot_from(self):
        self.settings.child('plot_options', 'plot_0d').setValue(dict(all_items=[], selected=[]))
        self.settings.child('plot_options', 'plot_1d').setValue(dict(all_items=[], selected=[]))

    def prepare_viewers(self):

        viewers_enum = [ViewersEnum('Data0D').increase_dim(self.scanner.n_axes)
                        for _ in range(len(self.settings['plot_options', 'plot_0d']['selected']))]
        data_names = self.settings['plot_options', 'plot_0d']['selected'][:]

        if self.settings['plot_options', 'group0D'] and len(viewers_enum) > 0 and ViewersEnum('Data1D') in viewers_enum:
            viewers_enum = [ViewersEnum('Data1D')]
            data_names = [self.live_plotter.grouped_data0D_fullname]
        elif self.settings['plot_options', 'group0D'] and len(viewers_enum) > 0 and ViewersEnum('Data2D') in viewers_enum:
            viewers_enum = [ViewersEnum('Data2D')]
            data_names = [self.live_plotter.grouped_data0D_fullname]

        if self.scanner.n_axes <= 1:
            viewers_enum.extend([ViewersEnum('Data1D').increase_dim(self.scanner.n_axes)
                                 for _ in range(len(self.settings['plot_options', 'plot_1d']['selected']))])
            data_names.extend(self.settings['plot_options', 'plot_1d']['selected'][:])
        self.live_plotter.prepare_viewers(viewers_enum, viewers_name=data_names)

    def update_status(self, txt, wait_time=0, log_type=None):
        """
            Show the txt message in the status bar with a delay of wait_time ms.

            =============== =========== =======================
            **Parameters**    **Type**    **Description**
            *txt*             string      The message to show
            *wait_time*       int         the delay of showing
            *log_type*        string      the type of the log
            =============== =========== =======================
        """
        self.ui.display_status(txt, wait_time)
        self.status_signal.emit(txt)
        logger.info(txt)

    @Slot(list)
    def thread_status(self, status):  # general function to get datas/infos from all threads back to the main
        """
            | General function to get datas/infos from all threads back to the main.
            |

            Switch the status with :
                * *"Update status"* : Update the status bar with the status attribute txt message
                * *"Update_scan_index"* : Set the value of the User Interface - indice_scan_sb attribute.
                * *"Scan_done"* : Save the scan and init the positions
                * *"Timeout"* : Set the "Timeout occured" in the User Interface-log message

            See Also
            --------
            update_status, save_scan, set_ini_positions
        """
        if status[0] == "Update_Status":
            self.update_status(status[1], wait_time=self.wait_time)

        elif status[0] == "Update_scan_index":
            # status[1] = [ind_scan,ind_average]
            self.ind_scan = status[1][0]
            self.ui.set_scan_step(status[1][0] + 1)
            self.ind_average = status[1][1]
            self.ui.set_scan_step_average(status[1][1] + 1)

        elif status[0] == "Scan_done":
            self.modules_manager.reset_signals()
            self.live_timer.stop()
            self.ui.set_scan_done()
            scan_node = self.module_and_data_saver.get_last_node()
            scan_node.attrs['scan_done'] = True
            self.module_and_data_saver.flush()

            if not self.batch_started:
                if not self.dashboard.overshoot:
                    self.set_ini_positions()
                self.ui.set_action_enabled('ini_positions', True)
                self.ui.set_action_enabled('start', True)

                # reactivate module controls usiong remote_control
                if hasattr(self.dashboard, 'remote_manager'):
                    remote_manager = getattr(self.dashboard, 'remote_manager')
                    remote_manager.activate_all(True)
                if self.navigator is not None:
                    self.navigator.list_2D_scans()
            else:
                self.ind_batch += 1
                self.loop_scan_batch()

        elif status[0] == "Timeout":
            self.ui.set_permanent_status('Timeout occurred')

    ############
    #  PLOTTING

    def save_temp_live_data(self, scan_data: ScanDataTemp):
        if scan_data.scan_index == 0:
            nav_axes = self.scanner.get_nav_axes()
            Naverage = self.settings['scan_options', 'scan_average']
            if Naverage > 1:
                for nav_axis in nav_axes:
                    nav_axis.index += 1
                nav_axes.append(data_mod.Axis('Average',
                                              data=np.linspace(0, Naverage - 1, Naverage),
                                              index=0))

            self.extended_saver.add_nav_axes(self.h5temp.raw_group, nav_axes)

        self.extended_saver.add_data(self.h5temp.raw_group, scan_data.data, scan_data.indexes,
                                     distribution=self.scanner.distribution)
        if self.settings['plot_options', 'plot_at_each_step']:
            self.update_live_plots()

    def update_live_plots(self):

        if self.settings['scan_options', 'scan_average'] > 1:
            average_axis = 0
        else:
            average_axis = None
        try:
            self.live_plotter.load_plot_data(group_0D=self.settings['plot_options', 'group0D'],
                                             average_axis=average_axis,
                                             average_index=self.ind_average,
                                             target_at=self.scanner.positions[self.ind_scan],
                                             last_step=(self.ind_scan ==
                                                        self.scanner.positions.size - 1 and
                                                        self.ind_average ==
                                                        self.settings[
                                                            'scan_options', 'scan_average'] - 1))
        except Exception as e:
            logger.exception(str(e))
    #################
    #  SCAN FLOW

    def set_scan(self, scan=None) -> bool:
        """
        Sets the current scan given the selected settings. Makes some checks, increments the h5 file scans.
        In case the dialog is cancelled, return False and aborts the scan
        """
        try:
            # set the filename and path
            if self.h5saver.h5_file is None:  # only the first time start scan is called
                self.create_new_file(True)
            res = self.update_scan_info()
            if not res:
                return False

            is_oversteps = self.scanner.set_scan()
            if is_oversteps:
                messagebox(text=f"An error occurred when establishing the scan steps. Actual settings "
                                f"gives approximately {int(self.scanner.n_steps)} steps."
                                f" Please check the steps number "
                                f"limit in the config file ({config['scan']['steps_limit']}) or modify"
                                f" your scan settings.")

            if self.modules_manager.Nactuators != self.scanner.n_axes:
                messagebox(text="There are not enough or too much selected move modules for this scan")
                return False

            if self.scanner.scan_sub_type == 'Adaptive':
                #todo include this in scanners objects for the adaptive scanners
                if len(self.modules_manager.get_selected_probed_data('0D')) == 0:
                    messagebox(text="In adaptive mode, you have to pick a 0D signal from which the algorithm will"
                                    " determine the next positions to scan, see 'probe_data' in the modules selector"
                                    " panel")
                    return False

            self.ui.n_scan_steps = self.scanner.n_steps

            # check if the modules are initialized
            for module in self.modules_manager.actuators:
                if not module.initialized_state:
                    raise exceptions.DAQ_ScanException('module ' + module.title + " is not initialized")

            for module in self.modules_manager.detectors:
                if not module.initialized_state:
                    raise exceptions.DAQ_ScanException('module ' + module.title + " is not initialized")

            self.ui.enable_start_stop(True)
            return True

        except Exception as e:
            logger.exception(str(e))
            self.ui.enable_start_stop(False)

    def set_metadata_about_current_scan(self):
        """
            Set the date/time and author values of the scan_info child of the scan_attributes tree.
            Show the 'scan' file attributes.

            See Also
            --------
            show_file_attributes
        """
        date = QDateTime(QDate.currentDate(), QTime.currentTime())
        self.scan_attributes.child('scan_info', 'date_time').setValue(date)
        self.scan_attributes.child('scan_info', 'author').setValue(
            self.dataset_attributes.child('dataset_info', 'author').value())
        if not self.batch_started:
            res = self.show_file_attributes('scan')
        else:
            res = True
        return res

    def set_metadata_about_dataset(self):
        """
            Set the date value of the data_set_info-date_time child of the data_set_attributes tree.
            Show the 'dataset' file attributes.

            See Also
            --------
            show_file_attributes
        """
        date = QDateTime(QDate.currentDate(), QTime.currentTime())
        self.dataset_attributes.child('dataset_info', 'date_time').setValue(date)
        res = self.show_file_attributes('dataset')
        self._metada_dataset_set = True
        return res

    def start_scan(self):
        """
            Start an acquisition calling the set_scan function.
            Emit the command_DAQ signal "start_acquisition".

            See Also
            --------
            set_scan
        """
        self.ui.display_status('Starting acquisition')
        self.dashboard.overshoot = False
        #deactivate double_clicked
        if self.ui.is_action_checked('move_at'):
            self.ui.get_action('move_at').trigger()

        res = self.set_scan()
        if res:
            # deactivate module controls using remote_control
            if hasattr(self.dashboard, 'remote_manager'):
                remote_manager = getattr(self.dashboard, 'remote_manager')
                remote_manager.activate_all(False)

            self.module_and_data_saver.h5saver = self.h5saver
            new_scan = self.module_and_data_saver.get_last_node().attrs['scan_done'] # get_last_node
            scan_node = self.module_and_data_saver.get_set_node(new=new_scan)
            self.save_metadata(scan_node, 'scan_info')

            self._init_live()

            # mandatory to deal with multithreads
            if self.scan_thread is not None:
                self.command_daq_signal.disconnect()
                if self.scan_thread.isRunning():
                    self.scan_thread.terminate()
                    while not self.scan_thread.isFinished():
                        QThread.msleep(100)
                    self.scan_thread = None

            self.scan_thread = QThread()

            scan_acquisition = DAQScanAcquisition(self.settings, self.scanner, self.h5saver.settings,
                                                  self.modules_manager,
                                                  module_saver=self.module_and_data_saver)
            if config['scan']['scan_in_thread']:
                scan_acquisition.moveToThread(self.scan_thread)
            self.command_daq_signal[utils.ThreadCommand].connect(scan_acquisition.queue_command)
            scan_acquisition.scan_data_tmp[ScanDataTemp].connect(self.save_temp_live_data)
            scan_acquisition.status_sig[list].connect(self.thread_status)

            self.scan_thread.scan_acquisition = scan_acquisition
            self.scan_thread.start()

            self.ui.set_action_enabled('ini_positions', False)
            self.ui.set_action_enabled('start', False)
            self.ui.set_scan_done(False)
            if not self.settings['plot_options', 'plot_at_each_step']:
                self.live_timer.start(self.settings['plot_options', 'refresh_live'])
            self.command_daq_signal.emit(utils.ThreadCommand('start_acquisition'))
            self.ui.set_permanent_status('Running acquisition')
            logger.info('Running acquisition')

    def _init_live(self):
        Naverage = self.settings['scan_options', 'scan_average']
        if Naverage > 1:
            scan_shape = [Naverage]
            scan_shape.extend(self.scanner.get_scan_shape())
        else:
            scan_shape = self.scanner.get_scan_shape()
        if self.temp_path is not None:
            try:
                self.h5temp.close()
                self.temp_path.cleanup()
            except Exception as e:
                logger.exception(str(e))

        self.h5temp = H5Saver()
        self.temp_path = tempfile.TemporaryDirectory(prefix='pymo')
        addhoc_file_path = Path(self.temp_path.name).joinpath('temp_data.h5')
        self.h5temp.init_file(custom_naming=True, addhoc_file_path=addhoc_file_path)
        self.extended_saver: data_saving.DataToExportExtendedSaver =\
            data_saving.DataToExportExtendedSaver(self.h5temp, extended_shape=scan_shape)
        self.live_plotter.h5saver = self.h5temp

        self.prepare_viewers()
        QtWidgets.QApplication.processEvents()

    def set_ini_positions(self):
        """
            Send the command_DAQ signal with "set_ini_positions" list item as an attribute.
        """
        self.command_daq_signal.emit(utils.ThreadCommand("set_ini_positions"))

    def stop_scan(self):
        """
            Emit the command_DAQ signal "stop_acquisition".

            See Also
            --------
            set_ini_positions
        """
        self.ui.set_permanent_status('Stoping acquisition')
        self.command_daq_signal.emit(utils.ThreadCommand("stop_acquisition"))
        scan_node = self.module_and_data_saver.get_last_node()
        scan_node.attrs['scan_done'] = True

        if not self.dashboard.overshoot:
            self.set_ini_positions()  # do not set ini position again in case overshoot fired
            status = 'Data Acquisition has been stopped by user'
        else:
            status = 'Data Acquisition has been stopped due to overshoot'

        self.update_status(status, log_type='log')
        self.ui.set_permanent_status('')

        self.ui.set_action_enabled('ini_positions', True)
        self.ui.set_action_enabled('start', True)

    def do_scan(self, start_scan=True):
        """Public method to start the scan programmatically"""
        if start_scan:
            if not self.ui.is_action_enabled('start'):
                self.ui.get_action('set_scan').trigger()
                QtWidgets.QApplication.processEvents()
            self.ui.get_action('start').trigger()
        else:
            self.ui.get_action('stop').trigger()


class DAQScanAcquisition(QObject):
    """
        =========================== ========================================

        =========================== ========================================

    """
    scan_data_tmp = Signal(ScanDataTemp)
    status_sig = Signal(list)

    def __init__(self, scan_settings: Parameter = None, scanner: Scanner = None,
                 h5saver_settings: Parameter = None, modules_manager: ModulesManager = None,
                 module_saver: module_saving.ScanSaver = None):

        """
        DAQScanAcquisition deal with the acquisition part of daq_scan, that is transferring commands to modules,
        getting back data, saviong and letting know th UI about the scan status

        """
        
        super().__init__()

        self.scan_settings = scan_settings
        self.modules_manager = modules_manager
        self.scanner = scanner

        self.stop_scan_flag = False
        self.Naverage = self.scan_settings['scan_options', 'scan_average']
        self.ind_average = 0
        self.ind_scan = 0

        self.isadaptive = self.scanner.scan_sub_type == 'Adaptive'

        self.modules_manager.timeout_signal.connect(self.timeout)
        self.timeout_scan_flag = False


        self.move_done_flag = False
        self.det_done_flag = False

        self.det_done_datas = data_mod.DataToExport('ScanData')

        self.h5saver = H5Saver()
        self.h5saver.settings.restoreState(h5saver_settings.saveState())
        self.h5saver.init_file(addhoc_file_path=self.h5saver.settings['current_h5_file'])

        self.module_and_data_saver: module_saving.ScanSaver = module_saver

        # update the DAQ_Viewer's detector saver to DetectorExtendedSaver to take into account extended
        # arrays due to scan shape and eventual averaging
        scan_shape = self.scanner.get_scan_shape()
        if self.Naverage > 1:
            self.scan_shape = [self.Naverage]
            self.scan_shape.extend(scan_shape)
        else:
            self.scan_shape = scan_shape

        for det in self.modules_manager.detectors:
            det.module_and_data_saver = module_saving.DetectorExtendedSaver(det, self.scan_shape)
        self.module_and_data_saver.h5saver = self.h5saver  # will update its h5saver and all submodules's h5saver

    @Slot(utils.ThreadCommand)
    def queue_command(self, command):
        """Process the commands sent by the main ui

        Parameters
        ----------
        command: utils.ThreadCommand
        """
        if command.command == "start_acquisition":
            self.start_acquisition()

        elif command.command == "stop_acquisition":
            self.stop_scan_flag = True

        elif command.command == "set_ini_positions":
            self.set_ini_positions()

        elif command.command == "move_stages":
            self.modules_manager.move_actuators(command.attribute)

    def set_ini_positions(self):
        """ Set the actuators's positions totheir initial value as defined in the scanner  """
        try:
            if self.scanner.scan_sub_type != 'Adaptive':
                self.modules_manager.move_actuators(self.scanner.positions_at(0))

        except Exception as e:
            logger.exception(str(e))

    def start_acquisition(self):
        try:
            #todo hoaw to apply newlayout to adaptive mode?

            self.modules_manager.connect_actuators()
            self.modules_manager.connect_detectors()

            self.stop_scan_flag = False

            Naxes = self.scanner.n_axes
            scan_type = self.scanner.scan_type
            self.navigation_axes = self.scanner.get_nav_axes()
            self.status_sig.emit(["Update_Status", "Acquisition has started", 'log'])

            self.timeout_scan_flag = False
            for ind_average in range(self.Naverage):
                self.ind_average = ind_average
                self.ind_scan = -1
                while True:
                    self.ind_scan += 1
                    if not self.isadaptive:
                        if self.ind_scan >= len(self.scanner.positions):
                            break
                        positions = self.scanner.positions_at(self.ind_scan)  # get positions
                    else:
                        pass
                        #todo update for v4
                        # positions = learner.ask(1)[0][-1]  # next point to probe
                        # if self.scanner.scan_type == 'Tabular':  # translate normalized curvilinear position to real coordinates
                        #     self.curvilinear = positions
                        #     length = 0.
                        #     for v in self.scanner.vectors:
                        #         length += v.norm()
                        #         if length >= self.curvilinear:
                        #             vec = v
                        #             frac_curvilinear = (self.curvilinear - (length - v.norm())) / v.norm()
                        #             break
                        #
                        #     position = (vec.vectorize() * frac_curvilinear).translate_to(vec.p1()).p2()
                        #     positions = [position.x(), position.y()]

                    self.status_sig.emit(["Update_scan_index", [self.ind_scan, ind_average]])

                    if self.stop_scan_flag or self.timeout_scan_flag:
                        break

                    #move motors of modules and wait for move completion
                    positions = self.modules_manager.order_positions(self.modules_manager.move_actuators(positions))

                    QThread.msleep(self.scan_settings['time_flow', 'wait_time_between'])

                    #grab datas and wait for grab completion
                    self.det_done(self.modules_manager.grab_datas(positions=positions), positions)

                    if self.isadaptive:
                        #todo update for v4
                        # det_channel = self.modules_manager.get_selected_probed_data()
                        # det, channel = det_channel[0].split('/')
                        # if self.scanner.scan_type == 'Tabular':
                        #     self.curvilinear_array.append(np.array([self.curvilinear]))
                        #     new_positions = self.curvilinear
                        # elif self.scanner.scan_type == 'Scan1D':
                        #     new_positions = positions[0]
                        # else:
                        #     new_positions = positions[:]
                        # learner.tell(new_positions, self.modules_manager.det_done_datas[det]['data0D'][channel]['data'])
                        pass

                    # daq_scan wait time
                    QThread.msleep(self.scan_settings.child('time_flow', 'wait_time').value())

            self.h5saver.flush()
            self.modules_manager.connect_actuators(False)
            self.modules_manager.connect_detectors(False)

            self.status_sig.emit(["Update_Status", "Acquisition has finished", 'log'])
            self.status_sig.emit(["Scan_done"])

        except Exception as e:
            logger.exception(str(e))
            # self.status_sig.emit(["Update_Status", getLineInfo() + str(e), 'log'])

    def det_done(self, det_done_datas: data_mod.DataToExport, positions):
        """

        """
        try:
            indexes = self.scanner.get_indexes_from_scan_index(self.ind_scan)
            if self.Naverage > 1:
                indexes = [self.ind_average] + list(indexes)
            indexes = tuple(indexes)
            if self.ind_scan == 0:
                nav_axes = self.scanner.get_nav_axes()
                if self.Naverage > 1:
                    for nav_axis in nav_axes:
                        nav_axis.index += 1
                    nav_axes.append(data_mod.Axis('Average', data=np.linspace(0, self.Naverage - 1,
                                                                              self.Naverage),
                                                  index=0))
                self.module_and_data_saver.add_nav_axes(nav_axes)

            self.module_and_data_saver.add_data(indexes=indexes, distribution=self.scanner.distribution)

            #todo related to adaptive (solution lies along the Enlargeable data saver)
            if self.isadaptive:
                for ind_ax, nav_axis in enumerate(self.navigation_axes):
                    nav_axis.append(np.array(positions[ind_ax]))

            self.det_done_flag = True

            full_names: list = self.scan_settings['plot_options', 'plot_0d']['selected'][:]
            full_names.extend(self.scan_settings['plot_options', 'plot_1d']['selected'][:])
            data_temp = det_done_datas.get_data_from_full_names(full_names, deepcopy=False)
            data_temp = data_temp.get_data_with_naxes_lower_than(2-len(indexes))  # maximum Data2D included nav indexes

            self.scan_data_tmp.emit(ScanDataTemp(self.ind_scan, indexes, data_temp))

        except Exception as e:
            logger.exception(str(e))
            # self.status_sig.emit(["Update_Status", getLineInfo() + str(e), 'log'])

    def timeout(self):
        """
            Send the status signal *'Time out during acquisition'*.
        """
        self.timeout_scan_flag = True
        self.status_sig.emit(["Update_Status", "Timeout during acquisition", 'log'])
        self.status_sig.emit(["Timeout"])



def main_test(init_qt=True):
    from pymodaq.utils.data import DataToExport

    LABEL = 'A Label'
    UNITS = 'units'
    OFFSET = -20.4
    SCALING = 0.22
    SIZE = 20
    DATA = OFFSET + SCALING * np.linspace(0, SIZE - 1, SIZE)

    DATA0D = np.array([2.7])
    DATA1D = np.arange(0, 10)
    DATA2D = np.arange(0, 5 * 6).reshape((5, 6))
    DATAND = np.arange(0, 5 * 6 * 3).reshape((5, 6, 3))

    def init_axis(data=None, index=0):
        if data is None:
            data = DATA
        return data_mod.Axis(label=LABEL, units=UNITS, data=data, index=index)

    def init_data(data=None, Ndata=1, axes=[], name='myData') -> data_mod.DataWithAxes:
        if data is None:
            data = DATA2D
        return data_mod.DataWithAxes(name, data_mod.DataSource(0), data=[data for ind in range(Ndata)],
                                     axes=axes)

    class ActuatorMock(QtCore.QObject):
        mod_name = 'act'
        move_done_signal = Signal(str, float)
        command_hardware = Signal(utils.ThreadCommand)

        def __init__(self, ind):
            super().__init__()
            self.title = f'{self.mod_name}_{ind:02d}'
            self.units = f'unit_{ind:02d}'
            self.initialized_state = True
            self.module_and_data_saver = module_saving.ActuatorSaver(self)
            self.settings = Parameter.create(name='settings', type='str', value='mysettings')
            self.ui = None
            self.command_hardware.connect(self.move_done)


        def move_done(self, command: utils.ThreadCommand):
            self.move_done_signal.emit(self.title, command.attribute[0])

    class DetectorMock(QtCore.QObject):
        mod_name = 'det'
        grab_done_signal = Signal(DataToExport)
        command_hardware = Signal(utils.ThreadCommand)

        def __init__(self, ind):
            super().__init__()
            self.title = f'{self.mod_name}_{ind:02d}'
            self.initialized_state = True
            self.module_and_data_saver = module_saving.DetectorSaver(self)
            self.settings = Parameter.create(name='settings', type='str', value='mysettings')
            self.ui = None
            self.command_hardware.connect(self.grab_done)

        def grab_done(self):
            dat1 = init_data(data=DATA2D, Ndata=2, name=f'{self.title}/data2D')
            dat2 = init_data(data=DATA1D, Ndata=3, name=f'{self.title}/data1D')
            data = data_mod.DataToExport(name=f'{self.title}', data=[dat1, dat2])
            self.grab_done_signal.emit(data)

    class DashBoardTest:
        def __init__(self):
            self.title = 'DashBoardTest'
            self.detector_modules = [DetectorMock(ind) for ind in range(2)]
            self.actuators_modules = [ActuatorMock(ind) for ind in range(3)]

    if init_qt:  # used for the test suite
        app = QtWidgets.QApplication(sys.argv)
        if config['style']['darkstyle']:
            import qdarkstyle
            app.setStyleSheet(qdarkstyle.load_stylesheet())

    win = QtWidgets.QMainWindow()
    area = gutils.dock.DockArea()
    win.setCentralWidget(area)
    #win.resize(1000, 500)
    win.setWindowTitle('PyMoDAQ Dashboard')

    dashboard = DashBoardTest()
    daq_scan = DAQScan(dockarea=area, dashboard=dashboard)
    win.show()

    if init_qt:
        sys.exit(app.exec_())
    return dashboard, daq_scan, win


def main(init_qt=True):
    if init_qt:  # used for the test suite
        app = QtWidgets.QApplication(sys.argv)
        if config['style']['darkstyle']:
            import qdarkstyle
            app.setStyleSheet(qdarkstyle.load_stylesheet())

    from pymodaq.dashboard import DashBoard

    win = QtWidgets.QMainWindow()
    area = gutils.dock.DockArea()
    win.setCentralWidget(area)
    win.resize(1000, 500)
    win.setWindowTitle('PyMoDAQ Dashboard')

    dashboard = DashBoard(area)
    daq_scan = None
    file = Path(get_set_preset_path()).joinpath(f"{config('presets', 'default_preset_for_scan')}.xml")
    if file.exists():
        dashboard.set_preset_mode(file)
        daq_scan = dashboard.load_scan_module()
    else:
        msgBox = QtWidgets.QMessageBox()
        msgBox.setText(f"The default file specified in the configuration file does not exists!\n"
                       f"{file}\n"
                       f"Impossible to load the DAQScan Module")
        msgBox.setStandardButtons(msgBox.Ok)
        ret = msgBox.exec()

    if init_qt:
        sys.exit(app.exec_())
    return dashboard, daq_scan, win


if __name__ == '__main__':
    main()
    #main_test()



from pymodaq.utils.daq_utils import ThreadCommand
from pymodaq.control_modules.move_utility_classes import DAQ_Move_base, comon_parameters_fun
from easydict import EasyDict as edict


class DAQ_Move_PID(DAQ_Move_base):
    """
    """
    _controller_units = 'whatever'
    is_multiaxes = True
    stage_names = []

    params = comon_parameters_fun(is_multiaxes, stage_names, master=False)

    def __init__(self, parent=None, params_state=None):
        super().__init__(parent, params_state)


    def update_position(self, dict_val):
        self.current_value = dict_val[self.parent.title]

    def get_actuator_value(self):
        self.controller['emit_curr_points'].emit()
        pos = self.current_value
        #
        # pos = self.get_position_with_scaling(pos)
        # self.current_value = pos
        return pos

    def close(self):
        pass

    def commit_settings(self, param):
        pass

    def ini_stage(self, controller=None):
        """
        """
        try:
            self.status.update(edict(info="", controller=None, initialized=False))
            if self.settings.child('multiaxes', 'ismultiaxes').value() and self.settings.child('multiaxes',
                                                                         'multi_status').value() == "Slave":
                if controller is None:
                    raise Exception('no controller has been defined externally while this axe is a slave one')
                else:
                    self.controller = controller
            else:  # Master stage
                self.controller = None  # any object that will control the stages

            self.controller['curr_point'].connect(self.update_position)

            info = "PID stage"
            self.status.info = info
            self.status.controller = self.controller
            self.status.initialized = True
            return self.status.info, self.status.initialized

        except Exception as e:
            self.status.info = str(e)
            self.status.initialized = False
            return self.status

    def move_Abs(self, position):
        """
        """
        position = self.check_bound(position)
        # position=self.set_position_with_scaling(position)
        # print(position)
        self.target_position = position

        self.controller['setpoint'].emit({self.parent.title: self.target_position})
        self.poll_moving()

    def move_Rel(self, position):
        """
        """
        position = self.check_bound(self.current_value + position) - self.current_value
        self.target_position = position + self.current_value

        self.controller['setpoint'].emit({self.parent.title: self.target_position})
        self.poll_moving()

    def move_Home(self):
        """
        """
        self.emit_status(ThreadCommand('Update_Status', ['Move Home not implemented']))

    def stop_motion(self):
        """
          Call the specific move_done function (depending on the hardware).

          See Also
          --------
          move_done
        """
        self.move_done()



import random

from pymodaq.utils.logger import set_logger, get_module_name
from pymodaq.utils import daq_utils as utils
from pymodaq.control_modules.move_utility_classes import params as daq_move_params
from pymodaq.control_modules.viewer_utility_classes import params as daq_viewer_params

from pyqtgraph.parametertree.Parameter import registerParameterType
from pyqtgraph.parametertree.parameterTypes.basetypes import GroupParameter
logger = set_logger(get_module_name(__file__))

DAQ_Move_Stage_type = utils.get_plugins('daq_move')
DAQ_0DViewer_Det_types = utils.get_plugins('daq_0Dviewer')
DAQ_1DViewer_Det_types = utils.get_plugins('daq_1Dviewer')
DAQ_2DViewer_Det_types = utils.get_plugins('daq_2Dviewer')
DAQ_NDViewer_Det_types = utils.get_plugins('daq_NDviewer')


def iterative_show_pb(params):
    for param in params:
        if param['type'] == 'itemselect' or param['type'] == 'list':
            param['show_pb'] = True
        elif 'children' in param:
            iterative_show_pb(param['children'])


class PresetScalableGroupMove(GroupParameter):
    """
        |

        ================ =============
        **Attributes**    **Type**
        *opts*            dictionnary
        ================ =============

        See Also
        --------
        hardware.DAQ_Move_Stage_type
    """

    def __init__(self, **opts):
        opts['type'] = 'groupmove'
        opts['addText'] = "Add"
        opts['addList'] = [mov['name'] for mov in DAQ_Move_Stage_type]
        super().__init__(**opts)

    def addNew(self, typ):
        """
            Add a child.

            =============== ===========
            **Parameters**   **Type**
            *typ*            string
            =============== ===========
        """
        name_prefix = 'move'

        child_indexes = [int(par.name()[len(name_prefix) + 1:]) for par in self.children()]

        if child_indexes == []:
            newindex = 0
        else:
            newindex = max(child_indexes) + 1

        params = daq_move_params
        iterative_show_pb(params)

        parent_module = utils.find_dict_in_list_from_key_val(DAQ_Move_Stage_type, 'name', typ)
        class_ = getattr(getattr(parent_module['module'], 'daq_move_' + typ),
                         'DAQ_Move_' + typ)
        params_hardware = getattr(class_, 'params')
        iterative_show_pb(params_hardware)

        for main_child in params:
            if main_child['name'] == 'move_settings':
                main_child['children'] = params_hardware
            elif main_child['name'] == 'main_settings':
                for child in main_child['children']:
                    if child['name'] == 'move_type':
                        child['value'] = typ
                    if child['name'] == 'controller_ID':
                        child['value'] = random.randint(0, 9999)

        child = {'title': 'Actuator {:02.0f}'.format(newindex), 'name': f'{name_prefix}{newindex:02.0f}',
                 'type': 'group',
                 'removable': True, 'children': [
                {'title': 'Name:', 'name': 'name', 'type': 'str', 'value': 'Move {:02.0f}'.format(newindex)},
                {'title': 'Init?:', 'name': 'init', 'type': 'bool', 'value': True},
                {'title': 'Settings:', 'name': 'params', 'type': 'group', 'children': params
                 }], 'removable': True, 'renamable': False}

        self.addChild(child)


registerParameterType('groupmove', PresetScalableGroupMove, override=True)


class PresetScalableGroupDet(GroupParameter):
    """
        =============== ==============
        **Attributes**    **Type**
        *opts*            dictionnary
        *options*         string list
        =============== ==============

        See Also
        --------
    """

    def __init__(self, **opts):
        opts['type'] = 'groupdet'
        opts['addText'] = "Add"
        options = []
        for name in [plugin['name'] for plugin in DAQ_0DViewer_Det_types]:
            options.append('DAQ0D/' + name)
        for name in [plugin['name'] for plugin in DAQ_1DViewer_Det_types]:
            options.append('DAQ1D/' + name)
        for name in [plugin['name'] for plugin in DAQ_2DViewer_Det_types]:
            options.append('DAQ2D/' + name)
        for name in [plugin['name'] for plugin in DAQ_NDViewer_Det_types]:
            options.append('DAQND/' + name)
        opts['addList'] = options

        super().__init__(**opts)

    def addNew(self, typ):
        """
            Add a child.

            =============== ===========  ================
            **Parameters**    **Type**   **Description*
            *typ*             string     the viewer name
            =============== ===========  ================
        """
        try:
            name_prefix = 'det'
            child_indexes = [int(par.name()[len(name_prefix) + 1:]) for par in self.children()]

            if child_indexes == []:
                newindex = 0
            else:
                newindex = max(child_indexes) + 1

            params = daq_viewer_params
            iterative_show_pb(params)

            for main_child in params:
                if main_child['name'] == 'main_settings':
                    for child in main_child['children']:
                        if child['name'] == 'DAQ_type':
                            child['value'] = typ[0:5]
                        if child['name'] == 'detector_type':
                            child['value'] = typ[6:]
                        if child['name'] == 'controller_status':
                            child['visible'] = True
                        if child['name'] == 'controller_ID':
                            child['value'] = random.randint(0, 9999)

            if '0D' in typ:
                parent_module = utils.find_dict_in_list_from_key_val(DAQ_0DViewer_Det_types, 'name', typ[6:])
                class_ = getattr(getattr(parent_module['module'], 'daq_0Dviewer_' + typ[6:]), 'DAQ_0DViewer_' + typ[6:])
            elif '1D' in typ:
                parent_module = utils.find_dict_in_list_from_key_val(DAQ_1DViewer_Det_types, 'name', typ[6:])
                class_ = getattr(getattr(parent_module['module'], 'daq_1Dviewer_' + typ[6:]), 'DAQ_1DViewer_' + typ[6:])
            elif '2D' in typ:
                parent_module = utils.find_dict_in_list_from_key_val(DAQ_2DViewer_Det_types, 'name', typ[6:])
                class_ = getattr(getattr(parent_module['module'], 'daq_2Dviewer_' + typ[6:]), 'DAQ_2DViewer_' + typ[6:])
            elif 'ND' in typ:
                parent_module = utils.find_dict_in_list_from_key_val(DAQ_NDViewer_Det_types, 'name', typ[6:])
                class_ = getattr(getattr(parent_module['module'], 'daq_NDviewer_' + typ[6:]), 'DAQ_NDViewer_' + typ[6:])
            for main_child in params:
                if main_child['name'] == 'main_settings':
                    for child in main_child['children']:
                        if child['name'] == 'axes':
                            child['visible'] = True

            params_hardware = getattr(class_, 'params')
            iterative_show_pb(params_hardware)

            for main_child in params:
                if main_child['name'] == 'detector_settings':
                    while len(main_child['children']) > 0:
                        for child in main_child['children']:
                            main_child['children'].remove(child)

                    main_child['children'].extend(params_hardware)

            child = {'title': 'Det {:02.0f}'.format(newindex), 'name': f'{name_prefix}{newindex:02.0f}',
                     'type': 'group', 'children': [
                {'title': 'Name:', 'name': 'name', 'type': 'str', 'value': 'Det {:02.0f}'.format(newindex)},
                {'title': 'Init?:', 'name': 'init', 'type': 'bool', 'value': True},
                {'title': 'Settings:', 'name': 'params', 'type': 'group', 'children': params},
            ], 'removable': True, 'renamable': False}

            self.addChild(child)
        except Exception as e:
            print(str(e))


registerParameterType('groupdet', PresetScalableGroupDet, override=True)



from abc import ABC, abstractmethod


class AbstractLogger(ABC):
    """
    Minimal set of methods a class should implement to be considered as a logger in PyMoDAQ and used by the DAQ_Logger
    extension

    See Also
    --------
    pymodaq.utils.h5modules.H5Logger or pymodaq.utils.db.db_logger.DbLoggerGUI
    """
    @abstractmethod
    def close(self):
        pass

    @property
    @abstractmethod
    def settings_tree(self):
        pass

    @property
    @abstractmethod
    def settings(self):
        pass

    @abstractmethod
    def init_logger(self, settings):
        pass

    @abstractmethod
    def get_handler(self):
        """
        returns a log handler to save the output of the error log into the particular implementation of the Logger object
        """
        pass

    @abstractmethod
    def add_detector(self, module_name, settings):
        pass

    @abstractmethod
    def add_actuator(self, module_name, settings):
        pass

    @abstractmethod
    def add_data(self, data):
        pass

    @abstractmethod
    def stop_logger(self):
        pass


# -*- coding: utf-8 -*-
"""
Created the 21/11/2022

@author: Sebastien Weber
"""
from time import time
from typing import Union, List, Tuple, Iterable

import numpy as np

from pymodaq.utils.abstract import ABCMeta, abstract_attribute
from pymodaq.utils.enums import enum_checker
from pymodaq.utils.data import Axis, DataDim, DataWithAxes, DataToExport, DataDistribution, DataDimError
from .saving import DataType, H5Saver
from .backends import GROUP, CARRAY, Node, EARRAY, NodeError
from pymodaq.utils.daq_utils import capitalize
from pymodaq.utils.scanner.utils import ScanType

SPECIAL_GROUP_NAMES = dict(nav_axes='NavAxes')


class AxisError(Exception):
    pass


class DataManagement(metaclass=ABCMeta):
    """Base abstract class to be used for all specialized object saving and loading data to/from a h5file

    Attributes
    ----------
    data_type: DataType
        The enum for this type of data, here abstract and should be redefined
    """
    data_type: DataType = abstract_attribute()
    _h5saver: H5Saver = abstract_attribute()

    @classmethod
    def _format_node_name(cls, ind: int) -> str:
        """ Format the saved node following the data_type attribute and an integer index

        Parameters
        ----------
        ind: int

        Returns
        -------
        str: the future name of the node
        """
        return f'{capitalize(cls.data_type.value)}{ind:02d}'

    def _get_next_node_name(self, where) -> str:
        """Get the formatted next node name given the ones already saved

        Parameters
        ----------
        where: Union[Node, str]
            the path of a given node or the node itself

        Returns
        -------
        str: the future name of the node
        """
        return self._format_node_name(self._get_next_data_type_index_in_group(where))

    def get_last_node_name(self, where) -> Union[str, None]:
        """Get the last node name among the ones already saved

        Parameters
        ----------
        where: Union[Node, str]
            the path of a given node or the node itself

        Returns
        -------
        str: the name of the last saved node or None if none saved
        """
        index = self._get_next_data_type_index_in_group(where) - 1
        if index == -1:
            return None
        else:
            return self._format_node_name(index)

    def get_node_from_index(self, where, index):
        return self._h5saver.get_node(where, self._format_node_name(index))

    def _get_next_data_type_index_in_group(self, where: Union[Node, str]) -> int:
        """Check how much node with a given data_type are already present within the GROUP where
        Parameters
        ----------
        where: Union[Node, str]
            the path of a given node or the node itself

        Returns
        -------
        int: the next available integer to index the node name
        """
        ind = 0
        for node in self._h5saver.walk_nodes(where):
            if 'data_type' in node.attrs:
                if node.attrs['data_type'] == self.data_type.name:
                    ind += 1
        return ind

    def _is_node_of_data_type(self, where):
        """Check if a given node is of the data_type of the real class implementation
        
        eg 'axis' for the AxisSaverLoader
        
        Parameters
        ----------
        where: Union[Node, str]
            the path of a given node or the node itself

        Returns
        -------
        bool
        """
        node = self._get_node(where)
        return 'data_type' in node.attrs and node.attrs['data_type'] == self.data_type

    def _get_node(self, where: Union[str, Node]) -> Node:
        """Utility method to get a node from a node or a string"""
        return self._h5saver.get_node(where)

    def _get_nodes(self, where: Union[str, Node]) -> List[Node]:
        """Get Nodes hanging from where including where

        Parameters
        ----------
        where: Union[Node, str]
            the path of a given node or the node itself

        Returns
        -------
        List[Node]
        """
        node = self._get_node(where)
        if isinstance(node, GROUP):
            return [child_node for child_node in self._h5saver.walk_nodes(node)]
        else:
            return [node]

    def _get_nodes_from_data_type(self, where):
        """Get the node list hanging from a parent and having the same data type as self
        
        Parameters
        ----------
        where: Union[Node, str]
            the path of a given node

        Returns
        -------
        list of Nodes
        """
        node = self._get_node(where)
        if isinstance(node, GROUP):
            parent_node = node
        else:
            parent_node = node.parent_node

        nodes = []
        for child_node in self._h5saver.walk_nodes(parent_node):
            if self._is_node_of_data_type(child_node):
                nodes.append(child_node)
        return nodes


class AxisSaverLoader(DataManagement):
    """Specialized Object to save and load Axis object to and from a h5file

    Parameters
    ----------
    h5saver: H5Saver

    Attributes
    ----------
    data_type: DataType
        The enum for this type of data, here 'axis'
    """
    data_type = DataType['axis']

    def __init__(self, h5saver: H5Saver):
        self._h5saver = h5saver
        self.data_type = enum_checker(DataType, self.data_type)

    def add_axis(self, where: Union[Node, str], axis: Axis, enlargeable=False):
        """Write Axis info at a given position within a h5 file

        Parameters
        ----------
        where: Union[Node, str]
            the path of a given node or the node itself
        axis: Axis
            the Axis object to add as a node in the h5file
        enlargeable: bool
            Specify if the underlying array will be enlargebale
        """
        array = self._h5saver.add_array(where, self._get_next_node_name(where), self.data_type, title=axis.label,
                                        array_to_save=axis.get_data(), data_dimension=DataDim['Data1D'],
                                        enlargeable=enlargeable,
                                        metadata=dict(size=axis.size, label=axis.label, units=axis.units,
                                                      index=axis.index, offset=axis.offset, scaling=axis.scaling,
                                                      distribution='uniform' if axis.is_axis_linear() else 'spread',
                                                      spread_order=axis.spread_order))
        return array

    def load_axis(self, where: Union[Node, str]) -> Axis:
        """create an Axis object from the data and metadata at a given node if of data_type: 'axis

        Parameters
        ----------
        where: Union[Node, str]
            the path of a given node or the node itself

        Returns
        -------
        Axis
        """
        axis_node = self._get_node(where)
        if not self._is_node_of_data_type(axis_node):
            raise AxisError(f'Could not create an Axis object from this node: {axis_node}')
        return Axis(label=axis_node.attrs['label'], units=axis_node.attrs['units'],
                    data=np.atleast_1d(np.squeeze(axis_node.read())), index=axis_node.attrs['index'],
                    spread_order=axis_node.attrs['spread_order'])

    def get_axes(self, where: Union[Node, str]) -> List[Axis]:
        """Return a list of Axis objects from the Axis Nodes hanging from (or among) a given Node

        Parameters
        ----------
        where: Union[Node, str]
            the path of a given node or the node itself

        Returns
        -------
        List[Axis]: the list of all Axis object
        """
        return [self.load_axis(node) for node in self._get_nodes_from_data_type(where)]


class DataSaverLoader(DataManagement):
    """Specialized Object to save and load DataWithAxes object to and from a h5file

    Parameters
    ----------
    h5saver: H5Saver

    Attributes
    ----------
    data_type: DataType
        The enum for this type of data, here 'data'
    """
    data_type = DataType['data']

    def __init__(self, h5saver: H5Saver):
        self.data_type = enum_checker(DataType, self.data_type)
        self._h5saver = h5saver
        self._axis_saver = AxisSaverLoader(h5saver)

    def isopen(self) -> bool:
        """ Get the opened status of the underlying hdf5 file"""
        return self._h5saver.isopen()

    def add_data(self, where: Union[Node, str], data: DataWithAxes, save_axes=True, **kwargs):
        """Adds Array nodes to a given location adding eventually axes as others nodes and metadata

        Parameters
        ----------
        where: Union[Node, str]
            the path of a given node or the node itself
        data: DataWithAxes
        save_axes: bool
        """

        for ind_data in range(len(data)):
            metadata = dict(timestamp=data.timestamp, label=data.labels[ind_data],
                            source=data.source.name, distribution=data.distribution.name,
                            origin=data.origin,
                            nav_indexes=tuple(data.nav_indexes)
                            if data.nav_indexes is not None else None)
            for name in data.extra_attributes:
                metadata[name] = getattr(data, name)
            self._h5saver.add_array(where, self._get_next_node_name(where), self.data_type, title=data.name,
                                    array_to_save=data[ind_data], data_dimension=data.dim.name, metadata=metadata)
        if save_axes:
            for axis in data.axes:
                self._axis_saver.add_axis(where, axis)

    def get_axes(self, where: Union[Node, str]) -> List[Axis]:
        """

        Parameters
        ----------
        where: Union[Node, str]
            the path of a given node or the node itself

        Returns
        -------

        """
        return self._axis_saver.get_axes(where)

    def get_bkg_nodes(self, where: Union[Node, str]):
        bkg_nodes = []
        for node in self._h5saver.walk_nodes(where):
            if 'data_type' in node.attrs and node.attrs['data_type'] == 'bkg':
                bkg_nodes.append(node)
        return bkg_nodes

    def get_data_arrays(self, where: Union[Node, str], with_bkg=False,
                        load_all=False) -> List[np.ndarray]:
        """

        Parameters
        ----------
        where: Union[Node, str]
            the path of a given node or the node itself
        with_bkg: bool
            If True try to load background node and return the array with background subtraction
        load_all: bool
            If True load all similar nodes hanging from a parent

        Returns
        -------
        list of ndarray
        """
        where = self._get_node(where)
        if with_bkg:
            bkg_nodes = []
            if with_bkg:
                bkg_nodes = self.get_bkg_nodes(where.parent_node)
            if len(bkg_nodes) == 0:
                with_bkg = False

        if load_all:
            getter = self._get_nodes_from_data_type
        else:
            getter = self._get_nodes

        if with_bkg:
            return [np.atleast_1d(np.squeeze(array.read()-bkg.read()))
                    for array, bkg in zip(getter(where), bkg_nodes)]
        else:
            return [np.atleast_1d(np.squeeze(array.read())) for array in getter(where)]

    def load_data(self, where, with_bkg=False, load_all=False) -> DataWithAxes:
        """Return a DataWithAxes object from the Data and Axis Nodes hanging from (or among) a
        given Node

        Does not include navigation axes stored elsewhere in the h5file. The node path is stored in
        the DatWithAxis using the attribute path

        Parameters
        ----------
        where: Union[Node, str]
            the path of a given node or the node itself
        with_bkg: bool
            If True try to load background node and return the data with background subtraction
        load_all: bool
            If True, will load all data hanging from the same parent node

        See Also
        --------
        load_data
        """

        data_node = self._get_node(where)

        if load_all:
            parent_node = data_node.parent_node
            data_nodes = self._get_nodes_from_data_type(parent_node)
            data_node = data_nodes[0]
        else:
            parent_node = data_node.parent_node
            if not isinstance(data_node, CARRAY):
                return
            data_nodes = [data_node]

        if 'axis' in self.data_type.name:
            ndarrays = [np.atleast_1d(np.squeeze(data_node.read())) for data_node in data_nodes]
            axes = [Axis(label=data_node.attrs['label'], units=data_node.attrs['units'],
                         data=np.linspace(0, ndarrays[0].size-1, ndarrays[0].size-1))]
        else:
            ndarrays = self.get_data_arrays(data_node, with_bkg=with_bkg, load_all=load_all)
            axes = self.get_axes(parent_node)

        extra_attributes = data_node.attrs.to_dict()
        for name in ['TITLE', 'CLASS', 'VERSION', 'backend', 'source', 'data_dimension',
                     'distribution', 'label', 'origin', 'nav_indexes', 'dtype', 'data_type',
                     'subdtype', 'shape', 'size', 'EXTDIM', 'path']:
            extra_attributes.pop(name, None)

        data = DataWithAxes(data_node.attrs['TITLE'],
                            source=data_node.attrs['source'] if 'source' in data_node.attrs
                            else 'raw',
                            dim=data_node.attrs['data_dimension'],
                            distribution=data_node.attrs['distribution'],
                            data=ndarrays,
                            labels=[node.attrs['label'] for node in data_nodes],
                            origin=data_node.attrs['origin'] if 'origin' in data_node.attrs else '',
                            nav_indexes=data_node.attrs['nav_indexes'] if 'nav_indexes' in
                                                                          data_node.attrs else (),
                            axes=axes,
                            path=data_node.path,
                            **extra_attributes)
        return data


class BkgSaver(DataSaverLoader):
    """Specialized Object to save and load DataWithAxes background object to and from a h5file

    Parameters
    ----------
    hsaver: H5Saver

    Attributes
    ----------
    data_type: DataType
        The enum for this type of data, here 'data'
    """
    data_type = DataType['bkg']

    def __init__(self, h5saver: H5Saver):
        super().__init__(h5saver)


class DataEnlargeableSaver(DataSaverLoader):
    """Specialized Object to save and load enlargeable DataWithAxes saved object to and from a h5file

    Particular case of DataND will a single *nav_indexes* parameter will be appended as chunks of signal data

    Parameters
    ----------
    h5saver: H5Saver

    Attributes
    ----------
    data_type: DataType
        The enum for this type of data, here 'data_enlargeable'

    Notes
    -----
    To be used to save data from a timed logger (DAQViewer continuous saving or DAQLogger extension) or from an
    adaptive scan where the final shape is unknown or other module that need this feature
    """
    data_type = DataType['data_enlargeable']

    def __init__(self, h5saver: H5Saver):
        super().__init__(h5saver)

    def _create_data_arrays(self, where: Union[Node, str], data: DataWithAxes, save_axes=True):
        """ Create enlargeable array to store data

        Parameters
        ----------
        where: Union[Node, str]
            the path of a given node or the node itself
        data: DataWithAxes
        save_axes: bool

        Notes
        -----
        Because data will be saved at a given index in the enlargeable array, related axes will have their index
        increased by one unity
        """

        if self.get_last_node_name(where) is None:
            for ind_data in range(len(data)):
                nav_indexes = list(data.nav_indexes)
                nav_indexes = [0] + list(np.array(nav_indexes, dtype=int) + 1)

                self._h5saver.add_array(where, self._get_next_node_name(where), self.data_type, title=data.name,
                                        array_to_save=data[ind_data],
                                        data_shape=data[ind_data].shape,
                                        array_type=data[ind_data].dtype,
                                        enlargeable=True,
                                        data_dimension=data.dim.name,
                                        metadata=dict(timestamp=data.timestamp, label=data.labels[ind_data],
                                                      source=data.source.name, distribution=data.distribution.name,
                                                      origin=data.origin,
                                                      nav_indexes=tuple(nav_indexes)))
            if save_axes:
                for axis in data.axes:
                    axis.index += 1  # because of enlargeable data will have an extra shape
                    self._axis_saver.add_axis(where, axis)

    def add_data(self, where: Union[Node, str], data: DataWithAxes):
        """ Append data to an enlargeable array node

        Data of dim (0, 1 or 2) will be just appended to the enlargeable array. DataND with one navigation axis of
        length (Lnav) will be considered as a collection of Lnav signal data of dim (0, 1 or 2) and will therefore be
        appended as Lnav signal data

        Parameters
        ----------
        where: Union[Node, str]
            the path of a given node or the node itself
        data: DataWithAxes
        """
        if self.get_last_node_name(where) is None:
            if len(data.nav_indexes) == 0:
                data_init = data
            elif len(data.nav_indexes) == 1:
                data_init = data.inav[0]
                data_init.source = data.source  # because slicing returns a calculated one
            else:
                raise DataDimError('It is not possible to append DataND with more than 1 navigation axis')
            self._create_data_arrays(where, data_init, save_axes=True)

        for ind_data in range(len(data)):
            array: EARRAY = self.get_node_from_index(where, ind_data)
            array.append(data[ind_data])


class DataExtendedSaver(DataSaverLoader):
    """Specialized Object to save and load DataWithAxes saved object to and from a h5file in extended arrays

    Parameters
    ----------
    h5saver: H5Saver
    extended_shape: Tuple[int]
        the extra shape compared to the data the h5array will have

    Attributes
    ----------
    data_type: DataType
        The enum for this type of data, here 'data'
    """
    data_type = DataType['data']

    def __init__(self, h5saver: H5Saver, extended_shape: Tuple[int]):
        super().__init__(h5saver)
        self.extended_shape = extended_shape

    def _create_data_arrays(self, where: Union[Node, str], data: DataWithAxes, save_axes=True,
                            distribution=DataDistribution['uniform']):
        """ Create array with extra dimensions (from scan) to store data

        Parameters
        ----------
        where: Union[Node, str]
            the path of a given node or the node itself
        data: DataWithAxes
        save_axes: bool

        Notes
        -----
        Because data will be saved at a given index in the "scan" array, related axes will have their index
        increased by the length of the scan dim (1 for scan1D, 2 for scan2D, ...)
        """
        if self.get_last_node_name(where) is None:
            for ind_data in range(len(data)):
                nav_indexes = list(data.nav_indexes)
                nav_indexes = [ind for ind in range(len(self.extended_shape))] +\
                              list(np.array(nav_indexes, dtype=int) + len(self.extended_shape))

                self._h5saver.add_array(where, self._get_next_node_name(where), self.data_type, title=data.name,
                                        data_shape=data[ind_data].shape,
                                        array_type=data[ind_data].dtype,
                                        scan_shape=self.extended_shape,
                                        add_scan_dim=True,
                                        data_dimension=data.dim.name,
                                        metadata=dict(timestamp=data.timestamp, label=data.labels[ind_data],
                                                      source=data.source.name, distribution=distribution.name,
                                                      origin=data.origin,
                                                      nav_indexes=tuple(nav_indexes)))

            if save_axes:
                for axis in data.axes:
                    axis.index += len(self.extended_shape)
                    # because there will be len(self.extended_shape) extra navigation axes
                    self._axis_saver.add_axis(where, axis)

    def add_data(self, where: Union[Node, str], data: DataWithAxes, indexes: List[int],
                 distribution=DataDistribution['uniform']):
        """Adds given DataWithAxes at a location within the initialized h5 array

        Parameters
        ----------
        where: Union[Node, str]
            the path of a given node or the node itself
        data: DataWithAxes
        indexes: Iterable[int]
            indexes where to save data in the init h5array (should have the same length as extended_shape and with values
            coherent with this shape
        """
        if len(indexes) != len(self.extended_shape):
            raise IndexError(f'Cannot put data into the h5array with extended indexes {indexes}')
        for ind in range(len(indexes)):
            if indexes[ind] > self.extended_shape[ind]:
                raise IndexError(f'Indexes cannot be higher than the array shape')

        if self.get_last_node_name(where) is None:
            self._create_data_arrays(where, data, save_axes=True, distribution=distribution)

        for ind_data in range(len(data)):
            #todo check that getting with index is safe...
            array: CARRAY = self.get_node_from_index(where, ind_data)
            array[tuple(indexes)] = data[ind_data]
            # maybe use array.__setitem__(indexes, data[ind_data]) if it's not working


class DataToExportSaver:
    """Object used to save DataToExport object into a h5file following the PyMoDAQ convention

    Parameters
    ----------
    h5saver: H5Saver

    """
    def __init__(self, h5saver: H5Saver):
        self._h5saver = h5saver
        self._data_saver = DataSaverLoader(h5saver)
        self._bkg_saver = BkgSaver(h5saver)

    def _get_node(self, where: Union[Node, str]) -> Node:
        return self._h5saver.get_node(where)

    def close(self):
        self._h5saver.close()

    def isopen(self) -> bool:
        """ Get the opened status of the underlying hdf5 file"""
        return self._h5saver.isopen()

    @staticmethod
    def channel_formatter(ind: int):
        """All DataWithAxes included in the DataToExport will be saved into a channel group indexed
        and formatted as below"""
        return f'CH{ind:02d}'

    def add_data(self, where: Union[Node, str], data: DataToExport, settings_as_xml='',
                 metadata=None):
        """

        Parameters
        ----------
        where: Union[Node, str]
            the path of a given node or the node itself
        data: DataToExport
        settings_as_xml: str
            The settings parameter as an XML string
        metadata: dict
            all extra metadata to be saved in the group node where data will be saved

        """
        if metadata is None:
            metadata = {}
        dims = data.get_dim_presents()
        for dim in dims:
            dim_group = self._h5saver.get_set_group(where, dim)
            for ind, dwa in enumerate(data.get_data_from_dim(dim)):
                # dwa: DataWithAxes filtered by dim
                dwa_group = self._h5saver.get_set_group(dim_group, self.channel_formatter(ind),
                                                        dwa.name)
                # dwa_group = self._h5saver.add_ch_group(dim_group, dwa.name)
                self._data_saver.add_data(dwa_group, dwa)

    def add_bkg(self, where: Union[Node, str], data: DataToExport):
        dims = data.get_dim_presents()
        for dim in dims:
            dim_group = self._h5saver.get_set_group(where, dim)
            for ind, dwa in enumerate(data.get_data_from_dim(dim)):
                # dwa: DataWithAxes filtered by dim
                dwa_group = self._h5saver.get_set_group(dim_group,
                                                        self.channel_formatter(ind), dwa.name)
                # dwa_group = self._get_node_from_title(dim_group, dwa.name)
                if dwa_group is not None:
                    self._bkg_saver.add_data(dwa_group, dwa, save_axes=False)


class DataToExportEnlargeableSaver(DataToExportSaver):
    """Generic object to save DataToExport objects in an enlargeable h5 array

    The next enlarged value should be specified in the add_data method

    Parameters
    ----------
    h5saver: H5Saver
    axis_name: str
        the name of the enlarged axis array
    axis_units: str
        the units of the enlarged axis array
    """
    def __init__(self, h5saver: H5Saver, axis_name: str = 'nav axis', axis_units: str = ''):

        super().__init__(h5saver)
        self._data_saver = DataEnlargeableSaver(h5saver)
        self._nav_axis_saver = AxisSaverLoader(h5saver)
        self._axis_name = axis_name
        self._axis_units = axis_units

    def add_data(self, where: Union[Node, str], data: DataToExport,
                 axis_value: Union[float, np.ndarray], settings_as_xml='', metadata=None):
        """

        Parameters
        ----------
        where: Union[Node, str]
            the path of a given node or the node itself
        data: DataToExport
            The data to be saved into an enlargeable array
        axis_value: float or np.ndarray
            The next value (or values) of the enlarged axis
        settings_as_xml: str
            The settings parameter as an XML string
        metadata: dict
            all extra metadata to be saved in the group node where data will be saved
        """
        super().add_data(where, data, settings_as_xml, metadata)
        where = self._get_node(where)
        nav_group = self._h5saver.get_set_group(where, SPECIAL_GROUP_NAMES['nav_axes'])
        if self._nav_axis_saver.get_last_node_name(nav_group) is None:
            axis = Axis(label=self._axis_name, units=self._axis_units, data=np.array([0., 1.]), index=0)
            axis_array = self._nav_axis_saver.add_axis(nav_group, axis, enlargeable=True)
            axis_array.attrs['size'] = 0

        axis_array = self._nav_axis_saver.get_node_from_index(nav_group, 0)
        axis_array.append(np.atleast_1d(np.squeeze(np.array([axis_value]))), expand=False)
        axis_array.attrs['size'] += 1


class DataToExportTimedSaver(DataToExportEnlargeableSaver):
    """Specialized DataToExportEnlargeableSaver to save data as a function of a time axis

    Only one element ca be added at a time, the time axis value are enlarged using the data to be
    added timestamp

    Notes
    -----
    This object is made for continuous saving mode of DAQViewer and logging to h5file for DAQLogger
    """
    def __init__(self, h5saver: H5Saver):
        super().__init__(h5saver, 'time', 's')

    def add_data(self, where: Union[Node, str], data: DataToExport, settings_as_xml='',
                 metadata=None):
        super().add_data(where, data, axis_value=data.timestamp, settings_as_xml=settings_as_xml,
                         metadata=metadata)


class DataToExportExtendedSaver(DataToExportSaver):
    """Object to save DataToExport at given indexes within arrays including extended shape

    Mostly used for data generated from the DAQScan

    Parameters
    ----------
    h5saver: H5Saver
    extended_shape: Tuple[int]
        the extra shape compared to the data the h5array will have
    """

    def __init__(self, h5saver: H5Saver, extended_shape: Tuple[int]):
        super().__init__(h5saver)
        self._data_saver = DataExtendedSaver(h5saver, extended_shape)
        self._nav_axis_saver = AxisSaverLoader(h5saver)

    def add_nav_axes(self, where: Union[Node, str], axes: List[Axis]):
        """Used to add navigation axes related to the extended array

        Notes
        -----
        For instance the scan axes in the DAQScan
        """
        where = self._get_node(where)
        nav_group = self._h5saver.get_set_group(where, SPECIAL_GROUP_NAMES['nav_axes'])
        if self._nav_axis_saver.get_last_node_name(nav_group) is None:
            for axis in axes:
                self._nav_axis_saver.add_axis(nav_group, axis)

    def add_data(self, where: Union[Node, str], data: DataToExport, indexes: Iterable[int],
                 distribution=DataDistribution['uniform'],
                 settings_as_xml='', metadata={}):
        """

        Parameters
        ----------
        where: Union[Node, str]
            the path of a given node or the node itself
        data: DataToExport
        indexes: List[int]
            indexes where to save data in the init h5array (should have the same length as
            extended_shape and with values coherent with this shape
        settings_as_xml: str
            The settings parameter as an XML string
        metadata: dict
            all extra metadata to be saved in the group node where data will be saved

        """
        dims = data.get_dim_presents()
        for dim in dims:
            dim_group = self._h5saver.get_set_group(where, dim)
            for ind, dwa in enumerate(data.get_data_from_dim(dim)):
                # dwa: DataWithAxes filtered by dim
                dwa_group = self._h5saver.get_set_group(dim_group,
                                                        self.channel_formatter(ind), dwa.name)
                self._data_saver.add_data(dwa_group, dwa, indexes=indexes,
                                          distribution=distribution)


class DataLoader:
    """Specialized Object to load DataWithAxes object from a h5file

    On the contrary to DataSaverLoader, does include navigation axes stored elsewhere in the h5file
    (for instance if saved from the DAQ_Scan)

    Parameters
    ----------
    h5saver: H5Saver
    """

    def __init__(self, h5saver: H5Saver):
        self._axis_loader: AxisSaverLoader = None
        self._data_loader: DataSaverLoader = None

        self.h5saver = h5saver

    @property
    def h5saver(self):
        return self._h5saver

    @h5saver.setter
    def h5saver(self, h5saver: H5Saver):
        self._h5saver = h5saver
        self._axis_loader = AxisSaverLoader(h5saver)
        self._data_loader = DataSaverLoader(h5saver)

    def get_node(self, where: Union[Node, str], name: str = None) -> Node:
        """ Convenience method to get node"""
        return self.h5saver.get_node(where, name)

    def get_nav_group(self, where: Union[Node, str]) -> Union[Node, None]:
        """

        Parameters
        ----------
        where: Union[Node, str]
            the path of a given node or the node itself

        Returns
        -------
        GROUP: returns the group named SPECIAL_GROUP_NAMES['nav_axes'] holding all NavAxis for
        those data

        See Also
        --------
        SPECIAL_GROUP_NAMES
        """
        node = self._h5saver.get_node(where)
        while node is not None:  # means we reached the root level
            if isinstance(node, GROUP):
                if self._h5saver.is_node_in_group(node, SPECIAL_GROUP_NAMES['nav_axes']):
                    return self._h5saver.get_node(node, SPECIAL_GROUP_NAMES['nav_axes'])
            node = node.parent_node

    def load_data(self, where: Union[Node, str], with_bkg=False, load_all=False) -> DataWithAxes:
        """Load data from a node (or channel node)

        Loaded data contains also nav_axes if any and with optional background subtraction

        Parameters
        ----------
        where: Union[Node, str]
            the path of a given node or the node itself
        with_bkg: bool
            If True will attempt to substract a background data node before loading
        load_all: bool
            If True, will load all data hanging from the same parent node

        Returns
        -------

        """
        node_data_type = DataType[self._h5saver.get_node(where).attrs['data_type']]
        self._data_loader.data_type = node_data_type
        data = self._data_loader.load_data(where, with_bkg=with_bkg, load_all=load_all)
        if 'axis' not in node_data_type.name:
            nav_group = self.get_nav_group(where)
            if nav_group is not None:
                nav_axes = self._axis_loader.get_axes(nav_group)
                data.axes.extend(nav_axes)
                data.get_dim_from_data_axes()
        data.create_missing_axes()
        return data

    def load_all(self, where: GROUP, data: DataToExport, with_bkg=False) -> DataToExport:

        where = self._h5saver.get_node(where)
        children_dict = where.children()
        data_list = []
        for child in children_dict:
            if isinstance(children_dict[child], GROUP):
                self.load_all(children_dict[child], data, with_bkg=with_bkg)
            elif ('data_type' in children_dict[child].attrs and 'data' in
                  children_dict[child].attrs['data_type']):

                data_list.append(self.load_data(children_dict[child].path,
                                                with_bkg=with_bkg, load_all=True))
                break
        data_tmp = DataToExport(name=where.name, data=data_list)
        data.append(data_tmp)



# -*- coding: utf-8 -*-
"""
Created the 04/11/2022

@author: Sebastien Weber
"""
from typing import Callable, Union
from abc import ABCMeta, abstractmethod

from pymodaq.utils.logger import set_logger, get_module_name

logger = set_logger(get_module_name(__file__))


class BuilderBase(ABCMeta):
    """Abstract class defining an object/service builder with a callable interface accepting some arguments

    See https://realpython.com/factory-method-python/ for some details

    See Also
    --------
    pymodaq.post_treatment.process_1d_to_scalar
    """
    @abstractmethod
    def __call__(self, *args, **kwargs):
        pass


class ObjectFactory(metaclass=ABCMeta):
    """Generic ObjectFactory with a decorator register to add object builders to the factory with a unique key
    identifier

    See https://realpython.com/factory-method-python/ for some details

    Examples
    --------
    @ObjectFactory.register('custom')
    def my_custom_builder():
        pass

    See Also
    --------
    pymodaq.post_treatment.process_1d_to_scalar.Data1DProcessorFactory
    """
    _builders = {}

    @classmethod
    def register(cls, key: str) -> Callable:
        def inner_wrapper(wrapped_class: Union[BuilderBase, Callable]) -> Callable:
            if cls.__name__ not in cls._builders:
                cls._builders[cls.__name__] = {}
            if key not in cls._builders[cls.__name__]:
                cls._builders[cls.__name__][key] = wrapped_class
            else:
                logger.warning(f'The {cls.__name__}/{key} builder is already registered. Replacing it')
            return wrapped_class
        return inner_wrapper

    @property
    def builders(self):
        return self._builders

    def keys_function(self, do_sort=True):
        if do_sort:
            return self.keys
        else:
            return list(self.builders[self.__class__.__name__].keys())

    @property
    def keys(self):
        return sorted(list(self.builders[self.__class__.__name__].keys()))

    @classmethod
    def create(cls, key, **kwargs):
        builder = cls._builders[cls.__name__].get(key)
        if not builder:
            raise ValueError(key)
        return builder(**kwargs)

    @classmethod
    def get_class(cls, key):
        return cls._builders[cls.__name__].get(key)



class DAQ_ScanException(Exception):
    """Raised when an error occur within the DAQScan"""
    pass


class DetectorError(Exception):
    pass


class ActuatorError(Exception):
    pass


class PIDError(Exception):
    pass


class MasterSlaveError(Exception):
    pass


class ViewerError(Exception):
    pass


class ExpectedError(Exception):
    """Raised in the tests made for pymodaq plugins"""
    pass


class Expected_1(ExpectedError):
    """Expected error 1 for pymodaq tests"""
    pass


class Expected_2(ExpectedError):
    """Expected error 2 for pymodaq tests"""
    pass


class Expected_3(ExpectedError):
    """Expected error 3 for pymodaq tests"""



from importlib import import_module
from pathlib import Path
from .utils import register_scanners


register_scanners()

from .scanner import Scanner  # import this one after the scanners because they have to first be registered



# -*- coding: utf-8 -*-
"""
Created the 06/01/2023

@author: Sebastien Weber
"""
import sys

from qtpy import QtSvg, QtWidgets
from qtpy.QtSvg import QSvgWidget


if __name__ == '__main__':
    app = QtWidgets.QApplication(sys.argv)

    widget = QSvgWidget()
    widget.show()
    widget.load(r'C:\Users\weber\Labo\Projet-Dossier candidature\Technical project\GDSII\wafer.svg')

    sys.exit(app.exec_())



from collections import OrderedDict
import datetime
import numpy as np

from pymodaq.utils.logger import set_logger, get_module_name
from pymodaq.utils.gui_utils.custom_app import CustomApp
from pymodaq.utils.gui_utils.dock import DockArea, Dock
from  pymodaq.utils.gui_utils.file_io import select_file
from pymodaq.utils.config import Config

from qtpy import QtWidgets
from qtpy.QtCore import Slot, QDate, QThread

from pymodaq.utils import daq_utils as utils
from pymodaq.utils.parameter import ioxml
from pymodaq.control_modules.daq_viewer import DAQ_Viewer
from pymodaq.utils.plotting.data_viewers.viewer0D import Viewer0D

from pymodaq.utils.h5modules.browsing import H5Browser
from pymodaq.utils.h5modules.saving import H5Saver
from pymodaq.utils.h5modules.data_saving import DataToExportSaver
from pymodaq.utils.data import DataToExport


config = Config()
logger = set_logger(get_module_name(__file__))


class CustomAppExample(CustomApp):

    # list of dicts enabling the settings tree on the user interface
    params = [
        {'title': 'Main settings:', 'name': 'main_settings', 'type': 'group', 'children': [
            {'title': 'Save base path:', 'name': 'base_path', 'type': 'browsepath',
             'value': config('data_saving', 'h5file', 'save_path')},
            {'title': 'File name:', 'name': 'target_filename', 'type': 'str', 'value': "", 'readonly': True},
            {'title': 'Date:', 'name': 'date', 'type': 'date', 'value': QDate.currentDate()},
            {'title': 'Do something, such as showing data:', 'name': 'do_something', 'type': 'bool', 'value': False},
            {'title': 'Something done:', 'name': 'something_done', 'type': 'led', 'value': False, 'readonly': True},
            {'title': 'Infos:', 'name': 'info', 'type': 'text', 'value': ""},
            {'title': 'push:', 'name': 'push', 'type': 'bool_push', 'value': False}
        ]},
        {'title': 'Other settings:', 'name': 'other_settings', 'type': 'group', 'children': [
            {'title': 'List of stuffs:', 'name': 'list_stuff', 'type': 'list', 'value': 'first',
                'limits': ['first', 'second', 'third'], 'tip': 'choose a stuff from the list'},
            {'title': 'List of integers:', 'name': 'list_int', 'type': 'list', 'value': 0,
                'limits': [0, 256, 512], 'tip': 'choose a stuff from this int list'},
            {'title': 'one integer:', 'name': 'an_integer', 'type': 'int', 'value': 500, },
            {'title': 'one float:', 'name': 'a_float', 'type': 'float', 'value': 2.7, },
        ]},
    ]

    def __init__(self, dockarea):

        super().__init__(dockarea)

        # init the object parameters
        self.raw_data: DataToExport = None
        self.setup_ui()

    def setup_actions(self):
        '''
        subclass method from ActionManager
        '''
        logger.debug('setting actions')
        self.add_action('quit', 'Quit', 'close2', "Quit program", toolbar=self.toolbar)
        self.add_action('grab', 'Grab', 'camera', "Grab from camera", checkable=True, toolbar=self.toolbar)
        self.add_action('load', 'Load', 'Open', "Load target file (.h5, .png, .jpg) or data from camera",
                        checkable=False, toolbar=self.toolbar)
        self.add_action('save', 'Save', 'SaveAs', "Save current data", checkable=False, toolbar=self.toolbar)
        self.add_action('show', 'Show/hide', 'read2', "Show Hide DAQViewer", checkable=True, toolbar=self.toolbar)

        logger.debug('actions set')

    def setup_docks(self):
        '''
        subclass method from CustomApp
        '''
        logger.debug('setting docks')
        self.dock_settings = Dock('Settings', size=(350, 350))
        self.dockarea.addDock(self.dock_settings, 'left')
        self.dock_settings.addWidget(self.settings_tree, 10)

        self.dock_logger = Dock("Logger")
        self.logger_list = QtWidgets.QListWidget()
        self.logger_list.setMinimumWidth(300)
        self.dock_logger.addWidget(self.logger_list)
        self.dockarea.addDock(self.dock_logger, 'bottom', self.dock_settings)

        # create a dock containing a viewer object, could be 0D, 1D or 2D depending what kind of data one want to plot here a 0D
        dock_Viewer0D = Dock('Viewer dock', size=(350, 350))
        self.dockarea.addDock(dock_Viewer0D, 'right', self.dock_logger)
        target_widget = QtWidgets.QWidget()
        self.target_viewer = Viewer0D(target_widget)
        dock_Viewer0D.addWidget(target_widget)

        # create 2 docks to display the DAQ_Viewer (one for its settings, one for its viewer)
        self.dock_detector_settings = Dock("Detector Settings", size=(350, 350))
        self.dockarea.addDock(self.dock_detector_settings, 'right', self.dock_settings)
        self.dock_detector = Dock("Detector Viewer", size=(350, 350))
        self.dockarea.addDock(self.dock_detector, 'right', self.dock_detector_settings)
        # init one daq_viewer object named detector

        self.detector = DAQ_Viewer(self.dockarea, dock_settings=self.dock_detector_settings,
                               dock_viewer=self.dock_detector, title="A detector")
        # set its type to 'Mock'
        self.detector.daq_type = 'DAQ0D'
        self.detector.detector = 'Mock'
        # init the detector and wait 1000ms for the completion
        self.detector.init_hardware()
        self.detector.settings.child('main_settings', 'wait_time').setValue(100)
        QtWidgets.QApplication.processEvents()
        QThread.msleep(1000)

        logger.debug('docks are set')

    def connect_things(self):
        '''
        subclass method from CustomApp
        '''
        logger.debug('connecting things')
        self.log_signal[str].connect(self.add_log)  # connect together this custom signal with the add_log method

        self.detector.grab_done_signal.connect(self.data_done)
        self.connect_action('quit', self.quit_function)
        self.connect_action('load', self.load_file)
        self.connect_action('save', self.save_data)

        self.connect_action('grab', self.detector.grab)
        self.connect_action('show', self.show_detector)
        logger.debug('connecting done')

    def show_detector(self, status):
        self.dock_detector.setVisible(status)
        self.dock_detector_settings.setVisible(status)

    def setup_menu(self):
        '''
        subclass method from CustomApp
        '''
        logger.debug('settings menu')
        file_menu = self.mainwindow.menuBar().addMenu('File')
        self.affect_to('quit', file_menu)
        file_menu.addSeparator()
        self.affect_to('load', file_menu)
        self.affect_to('save', file_menu)

        self.affect_to('quit', file_menu)

        logger.debug('menu set')

    def value_changed(self, param):
        logger.debug(f'calling value_changed with param {param.name()}')
        if param.name() == 'do_something':
            if param.value():
                self.log_signal.emit('Do something')
                self.detector.grab_done_signal.connect(self.show_data)
                self.raw_data = []  # init the data to be finally saved
                self.settings.child('main_settings', 'something_done').setValue(True)
            else:
                self.log_signal.emit('Stop Doing something')
                self.detector.grab_done_signal.disconnect()
                self.settings.child('main_settings', 'something_done').setValue(False)

        logger.debug(f'Value change applied')

    def data_done(self, data):
        # print(data)
        pass

    def show_data(self, data: DataToExport):
        """
        do stuff with data from the detector if its grab_done_signal has been connected
        Parameters
        ----------
        data: DataToExport
        """
        self.raw_data = data
        data0D = data.get_data_from_dim('Data0D')

        self.target_viewer.show_data(data0D.data[0])

    def load_file(self):
        # init the data browser module
        widg = QtWidgets.QWidget()
        self.data_browser = H5Browser(widg)
        widg.show()

    def quit_function(self):
        # close all stuff that need to be
        self.detector.quit_fun()
        QtWidgets.QApplication.processEvents()
        self.mainwindow.close()

    def run_detector(self):
        self.detector.ui.grab_pb.click()

    def save_data(self):
        try:
            path = select_file(start_path=self.settings.child('main_settings', 'base_path').value(), save=True,
                                                                   ext='h5')
            if path is not None:
                # init the file object with an addhoc name given by the user
                h5saver = H5Saver(save_type='custom')
                h5saver.init_file(update_h5=True, addhoc_file_path=path)
                datasaver = DataToExportSaver(h5saver)


                # save all metadata
                settings_str = ioxml.parameter_to_xml_string(self.settings)
                settings_str = b'<All_settings>' + settings_str
                settings_str += ioxml.parameter_to_xml_string(self.detector.settings) + ioxml.parameter_to_xml_string(
                    h5saver.settings) + b'</All_settings>'

                datasaver.add_data(h5saver.root(), self.raw_data, settings_as_xml=settings_str)

                h5saver.close_file()

                st = 'file {:s} has been saved'.format(str(path))
                self.add_log(st)
                self.settings.child('main_settings', 'info').setValue(st)

        except Exception as e:
            logger.exception(str(e))

    @Slot(str)
    def add_log(self, txt):
        """
            Add a log to the logger list from the given text log and the current time

            ================ ========= ======================
            **Parameters**   **Type**   **Description**

             *txt*             string    the log to be added
            ================ ========= ======================

        """
        now = datetime.datetime.now()
        new_item = QtWidgets.QListWidgetItem(str(now) + ": " + txt)
        self.logger_list.addItem(new_item)
        logger.info(txt)


def main():
    import sys
    app = QtWidgets.QApplication(sys.argv)
    mainwindow = QtWidgets.QMainWindow()
    dockarea = DockArea()
    mainwindow.setCentralWidget(dockarea)

    prog = CustomAppExample(dockarea)

    mainwindow.show()
    sys.exit(app.exec_())


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
"""
Created the 02/03/2023

@author: Sebastien Weber
"""
import numpy as np


# project imports
from pymodaq.utils.h5modules.backends import Node
from pymodaq.utils.h5modules.exporter import ExporterFactory, H5Exporter


@ExporterFactory.register_exporter()
class H5asciiExporter(H5Exporter):
    """ Exporter object for saving nodes as txt files"""

    FORMAT_DESCRIPTION = "Ascii flimj file"
    FORMAT_EXTENSION = "ascii"

    def export_data(self, node: Node, filename: str) -> None:
        if 'ARRAY' in node.attrs['CLASS']:
            data = node.read()
            if not isinstance(data, np.ndarray):
                # in case one has a list of same objects (array of strings for instance, logger or other)
                data = np.array(data)
                np.savetxt(filename,
                           data.T if len(data.shape) > 1 else [data],
                           '%s', '\t')
            else:
                np.savetxt(filename,
                           data.T if len(data.shape) > 1 else [data],
                           '%.6e', '\t')

        elif 'GROUP' in node.attrs['CLASS']:
            data_tot = []
            header = []
            dtypes = []
            fmts = []
            for subnode_name, subnode in node.children().items():
                if 'ARRAY' in subnode.attrs['CLASS']:
                    if len(subnode.attrs['shape']) == 1:
                        data = subnode.read()
                        if not isinstance(data, np.ndarray):
                            # in case one has a list of same objects (array of strings for instance, logger or other)
                            data = np.array(data)
                        data_tot.append(data)
                        dtypes.append((subnode_name, data.dtype))
                        header.append(subnode_name)
                        if data.dtype.char == 'U':
                            fmt = '%s'  # for strings
                        elif data.dtype.char == 'l':
                            fmt = '%d'  # for integers
                        else:
                            fmt = '%.6f'  # for decimal numbers
                        fmts.append(fmt)

            data_trans = np.array(list(zip(*data_tot)), dtype=dtypes)

            np.savetxt(filename, data_trans, fmts, '\t', header='#' + '\t'.join(header))





from typing import List

import numpy as np

from qtpy import QtWidgets
from qtpy.QtCore import QObject
from pymodaq.utils.plotting.data_viewers.viewer0D import Viewer0D
from pymodaq.utils.data import DataRaw
import sys


class LCD(QObject):

    def __init__(self, parent, **kwargs):
        super().__init__()
        self.Nvals = None
        self.labels = None
        self.parent = parent
        self.viewer0D = None
        self.digits = 3
        self.setupui(**kwargs)

    def setvalues(self, values: List[np.ndarray]):
        """
        display values on lcds
        Parameters
        ----------
        values: list of 0D ndarray

        Returns
        -------

        """
        while len(values) < self.Nvals:
            values.append(np.array([0.]))
        if len(values) > self.Nvals:
            values = values[:self.Nvals]
        vals = []
        for ind, val in enumerate(values):
            self.lcds[ind].display(val[0])
            vals.append(val)
        self.viewer0D.show_data(DataRaw(name='LCD', data=values))

    def setupui(self, **kwargs):
        if 'digits' in kwargs:
            self.digits = kwargs['digits']
        if 'Nvals' in kwargs:
            self.Nvals = kwargs['Nvals']
        else:
            self.Nvals = 1
        if 'labels' in kwargs:
            self.labels = kwargs['labels']
        else:
            self.labels = ['CH{:d}'.format(ind) for ind in range(self.Nvals)]

        while len(self.labels) < self.Nvals:
            self.labels.append('')

        vlayout = QtWidgets.QVBoxLayout()
        hsplitter = QtWidgets.QSplitter()
        vlayout.addWidget(hsplitter)
        self.parent.setLayout(vlayout)
        form = QtWidgets.QWidget()
        self.viewer0D = Viewer0D(form)
        self.viewer0D.labels = self.labels

        vlayout = QtWidgets.QVBoxLayout()

        lcd_layouts = []
        labels = []
        self.lcds = []

        for ind in range(self.Nvals):
            lcd_layouts.append(QtWidgets.QVBoxLayout())
            labels.append(QtWidgets.QLabel(self.labels[ind]))
            labels[-1].setSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Maximum)
            self.lcds.append(QtWidgets.QLCDNumber())
            self.lcds[-1].setDigitCount(self.digits)
            lcd_layouts[-1].addWidget(labels[-1])
            lcd_layouts[-1].addWidget(self.lcds[-1])
            vlayout.addLayout(lcd_layouts[-1])

            if ind != self.Nvals - 1:
                hFrame = QtWidgets.QFrame()
                hFrame.setFrameShape(QtWidgets.QFrame.HLine)
                vlayout.addWidget(hFrame)

        lcd_widget = QtWidgets.QWidget()
        lcd_widget.setLayout(vlayout)
        hsplitter.addWidget(lcd_widget)
        hsplitter.addWidget(form)
        self.parent.resize(800, 500)
        hsplitter.setSizes([400, 300])


if __name__ == '__main__':
    from pymodaq.utils.math_utils import gauss1D
    import numpy as np

    x = np.linspace(0, 200, 201)
    y1 = gauss1D(x, 75, 25)
    y2 = gauss1D(x, 120, 50, 2)
    app = QtWidgets.QApplication(sys.argv)
    Form = QtWidgets.QWidget()

    prog = LCD(Form, Nvals=2)
    Form.show()
    for ind, data in enumerate(y1):
        prog.setvalues([data])
        QtWidgets.QApplication.processEvents()
    sys.exit(app.exec_())



from collections.abc import Iterable

import copy
from numbers import Real, Number
from typing import List, Union
from typing import Iterable as IterableType

from easydict import EasyDict as edict
from multipledispatch import dispatch
import numpy as np
import pyqtgraph as pg
from qtpy import QtGui, QtCore, QtWidgets
from scipy.spatial import Delaunay as Triangulation

from pymodaq.utils import data as data_mod
from pymodaq.utils.plotting.items.axis_scaled import AxisItem_Scaled
from pymodaq.utils.plotting.data_viewers.viewer1Dbasic import Viewer1DBasic
from pymodaq.utils import daq_utils as utils
from pymodaq.utils.messenger import deprecation_msg


class Point:
    def __init__(self, *elt: IterableType[float]):
        """Initialize a geometric point in an arbitrary number of dimensions

        Parameters
        ----------
        elt: either a tuple of floats, passed as multiple parameters or a single Iterable parameter
        """
        if len(elt) == 1 and isinstance(elt[0], Iterable):
            elt = elt[0]

        self._coordinates = np.atleast_1d(np.squeeze(elt))
        self._ndim = len(elt)

    @property
    def coordinates(self):
        return self._coordinates

    def copy(self):
        return Point(self.coordinates.copy())

    def __getitem__(self, item: int):
        return self._coordinates[item]

    def __setitem__(self, key: int, value: float):
        self._coordinates[key] = value

    def __len__(self):
        return self._ndim

    def _compare_length(self, other: 'Point'):
        if len(self) != len(other):
            raise ValueError('Those points should be expressed in the same coordinate system and dimensions')

    def __add__(self, other: Union['Point', 'Vector']):
        self._compare_length(other)
        return Point(*(self._coordinates + other._coordinates))

    def __sub__(self, other: 'Point'):
        self._compare_length(other)
        return Point(*(self._coordinates - other._coordinates))

    def __repr__(self):
        return f'Point({self.coordinates})'


class Vector:
    def __init__(self, coordinates: Union[Point, np.ndarray], origin: Point = None):
        if isinstance(coordinates, Point):
            self._coordinates = coordinates.coordinates
        else:
            self._coordinates = coordinates

        if origin is None:
            origin = np.zeros((len(coordinates)))
        else:
            self._compare_length(origin)

        self._origin = origin

    def _compare_length(self, other: 'Point'):
        if len(self) != len(other):
            raise ValueError('Those Points/Vectors should be expressed in the same coordinate system and dimensions')

    @property
    def origin(self):
        return self._origin

    @property
    def coordinates(self):
        return self._coordinates

    def copy(self):
        return Vector(self.coordinates.copy(), origin=self.origin.copy())

    def __len__(self):
        return len(self._coordinates)

    def norm(self):
        return np.linalg.norm(self._coordinates)

    def unit_vector(self):
        return self * (1 / self.norm())

    def __add__(self, other: 'Vector'):
        self._compare_length(other)
        return Vector(self.coordinates + other.coordinates, origin=self.origin.copy())

    def __sub__(self, other: 'Vector'):
        self._compare_length(other)
        return Vector(self.coordinates - other.coordinates, origin=self.origin.copy())

    def __mul__(self, other: Number):
        if not isinstance(other, Number):
            raise TypeError(f'Cannot multiply a vector with {other}')
        return Vector(other * self.coordinates, origin=self.origin.copy())

    def dot(self, other: 'Vector'):
        self._compare_length(other)
        return np.dot(self.coordinates, other.coordinates)

    def cross(self, other: 'Vector'):
        self._compare_length(other)
        return np.cross(self.coordinates, other.coordinates)

    def __repr__(self):
        return f'Vector({self.coordinates})/Origin({self.origin.coordinates})'


def get_sub_segmented_positions(spacing: float, points: List[Point]) -> List[np.ndarray]:
    """Get Points coordinates spaced in between subsequent Points

    Parameters
    ----------
    spacing: float
        Distance between two subpoints
    points: List[Point]
        List of Points in arbitrary dimension forming segments one want to sample with a distance equal to spacing

    Returns
    -------
    List[np.ndarray]: The list of the coordinates of the points
    """
    positions = []
    for ind in range(len(points) - 1):
        vect = Vector(points[ind+1]-points[ind], origin=points[ind])
        npts = 0
        while npts * spacing < vect.norm():
            positions.append(
                (vect.origin + vect.unit_vector() * npts * spacing).coordinates)
            npts += 1

    # # add_last point not taken into account
    positions.append(points[-1].coordinates)
    return positions


class QVector(QtCore.QLineF):
    def __init__(self, *elt):
        super().__init__(*elt)

    def __repr__(self):
        return f"PyMoDAQ's QVector({self.x1()}, {self.y1()}, {self.x2()}, {self.y2()})"

    def __add__(self, qvect):
        v = QVector(self.x1() + qvect.x1(), self.y1() + qvect.y1(), self.x2() + qvect.x2(), self.y2() + qvect.y2())
        return v

    def __sub__(self, qvect):
        v = QVector(self.x1() - qvect.x1(), self.y1() - qvect.y1(), self.x2() - qvect.x2(), self.y2() - qvect.y2())
        return v

    def __mul__(self, coeff=float(1)):
        v = QVector(coeff * self.x1(), coeff * self.y1(), coeff * self.x2(), coeff * self.y2())
        return v

    def copy(self):
        vec = QVector()
        vec.setPoints(copy.copy(self.p1()), copy.copy(self.p2()))
        return vec

    def vectorize(self):
        v = QVector(QtCore.QPointF(0, 0), self.p2() - self.p1())
        return v

    def norm(self):
        return self.length()

    def unitVector(self):
        vec = self * (1 / self.length())
        return vec

    def normalVector(self):
        vec = self.vectorize()
        vec = QVector(0, 0, -vec.p2().y(), vec.p2().x())
        return vec

    def normalVector_not_vectorized(self):
        vec = self.vectorize()
        vec = QVector(0, 0, -vec.p2().y(), vec.p2().x())
        vec.translate(self.p1())
        return vec

    def dot(self, qvect):
        """
        scalar product
        """
        v1 = self.vectorize()
        v2 = qvect.vectorize()
        prod = v1.x2() * v2.x2() + v1.y2() * v2.y2()
        return prod

    def prod(self, qvect):
        """
        vectoriel product length along z
        """
        v1 = self.vectorize()
        v2 = qvect.vectorize()
        prod = v1.x2() * v2.y2() - v1.y2() * v2.x2()
        return prod

    def translate_to(self, point=QtCore.QPointF(0, 0)):
        vec = self + QVector(self.p1(), point)
        return vec


def makeAlphaTriangles(data, lut=None, levels=None, scale=None, useRGBA=False):
    """
    Convert an array of values into an ARGB array suitable for building QImages,
    OpenGL textures, etc.

    Returns the ARGB array (unsigned byte) and a boolean indicating whether
    there is alpha channel data. This is a two stage process:
        0) compute the polygons (triangles) from triangulation of the points
        1) Rescale the data based on the values in the *levels* argument (min, max).
        2) Determine the final output by passing the rescaled values through a
           lookup table.

    Both stages are optional.

    ============== ==================================================================================
    **Arguments:**
    data           numpy array of int/float types. If
    levels         List [min, max]; optionally rescale data before converting through the
                   lookup table. The data is rescaled such that min->0 and max->*scale*::

                      rescaled = (clip(data, min, max) - min) * (*scale* / (max - min))

                   It is also possible to use a 2D (N,2) array of values for levels. In this case,
                   it is assumed that each pair of min,max values in the levels array should be
                   applied to a different subset of the input data (for example, the input data may
                   already have RGB values and the levels are used to independently scale each
                   channel). The use of this feature requires that levels.shape[0] == data.shape[-1].
    scale          The maximum value to which data will be rescaled before being passed through the
                   lookup table (or returned if there is no lookup table). By default this will
                   be set to the length of the lookup table, or 255 if no lookup table is provided.
    lut            Optional lookup table (array with dtype=ubyte).
                   Values in data will be converted to color by indexing directly from lut.
                   The output data shape will be input.shape + lut.shape[1:].
                   Lookup tables can be built using ColorMap or GradientWidget.
    useRGBA        If True, the data is returned in RGBA order (useful for building OpenGL textures).
                   The default is False, which returns in ARGB order for use with QImage
                   (Note that 'ARGB' is a term used by the Qt documentation; the *actual* order
                   is BGRA).
    ============== ==================================================================================
    """
    points = data[:, :2]
    values = data[:, 2]

    profile = pg.debug.Profiler()
    if points.ndim not in (2,):
        raise TypeError("points must be 1D sequence of points")

    tri = Triangulation(points)
    tri_data = np.zeros((len(tri.simplices),))
    for ind, pts in enumerate(tri.simplices):
        tri_data[ind] = np.mean(values[pts])
    data = tri_data.copy()
    if lut is not None and not isinstance(lut, np.ndarray):
        lut = np.array(lut)

    if levels is None:
        # automatically decide levels based on data dtype
        if data.dtype.kind == 'u':
            levels = np.array([0, 2 ** (data.itemsize * 8) - 1])
        elif data.dtype.kind == 'i':
            s = 2 ** (data.itemsize * 8 - 1)
            levels = np.array([-s, s - 1])
        elif data.dtype.kind == 'b':
            levels = np.array([0, 1])
        else:
            raise Exception('levels argument is required for float input types')
    if not isinstance(levels, np.ndarray):
        levels = np.array(levels)
    if levels.ndim == 1:
        if levels.shape[0] != 2:
            raise Exception('levels argument must have length 2')
    elif levels.ndim == 2:
        if lut is not None and lut.ndim > 1:
            raise Exception('Cannot make ARGB data when both levels and lut have ndim > 2')
        if levels.shape != (data.shape[-1], 2):
            raise Exception('levels must have shape (data.shape[-1], 2)')
    else:
        raise Exception("levels argument must be 1D or 2D (got shape=%s)." % repr(levels.shape))

    profile()

    # Decide on maximum scaled value
    if scale is None:
        if lut is not None:
            scale = lut.shape[0] - 1
        else:
            scale = 255.

    # Decide on the dtype we want after scaling
    if lut is None:
        dtype = np.ubyte
    else:
        dtype = np.min_scalar_type(lut.shape[0] - 1)

    # Apply levels if given
    if levels is not None:
        if isinstance(levels, np.ndarray) and levels.ndim == 2:
            # we are going to rescale each channel independently
            if levels.shape[0] != data.shape[-1]:
                raise Exception(
                    "When rescaling multi-channel data, there must be the same number of levels as channels (data.shape[-1] == levels.shape[0])")
            newData = np.empty(data.shape, dtype=int)
            for i in range(data.shape[-1]):
                minVal, maxVal = levels[i]
                if minVal == maxVal:
                    maxVal += 1e-16
                newData[..., i] = pg.functions.rescaleData(data[..., i], scale / (maxVal - minVal), minVal, dtype=dtype)
            data = newData
        else:
            # Apply level scaling unless it would have no effect on the data
            minVal, maxVal = levels
            if minVal != 0 or maxVal != scale:
                if minVal == maxVal:
                    maxVal += 1e-16
                data = pg.functions.rescaleData(data, scale / (maxVal - minVal), minVal, dtype=dtype)

    profile()

    # apply LUT if given
    if lut is not None:
        data = pg.functions.applyLookupTable(data, lut)
    else:
        if data.dtype is not np.ubyte:
            data = np.clip(data, 0, 255).astype(np.ubyte)

    profile()

    # this will be the final image array
    imgData = np.empty((data.shape[0],) + (4,), dtype=np.ubyte)

    profile()

    # decide channel order
    if useRGBA:
        order = [0, 1, 2, 3]  # array comes out RGBA
    else:
        order = [2, 1, 0, 3]  # for some reason, the colors line up as BGR in the final image.

    # TODO check this
    # copy data into image array
    if data.ndim == 1:
        # This is tempting:
        #   imgData[..., :3] = data[..., np.newaxis]
        # ..but it turns out this is faster:
        for i in range(3):
            imgData[..., i] = data
    elif data.shape[1] == 1:
        for i in range(3):
            imgData[..., i] = data[..., 0]
    else:
        for i in range(0, data.shape[1]):
            imgData[..., i] = data[..., order[i]]

    profile()

    # add opaque alpha channel if needed
    if data.ndim == 1 or data.shape[1] == 3:
        alpha = False
        imgData[..., 3] = 255
    else:
        alpha = True

    profile()
    return tri, tri_data, imgData, alpha


def makePolygons(tri):
    polygons = []
    for seq in tri.points[tri.simplices]:
        polygons.append(QtGui.QPolygonF([QtCore.QPointF(*s) for s in seq] + [QtCore.QPointF(*seq[0])]))
    return polygons


class Data0DWithHistory:
    """Object to store scalar values and keep a history of a given length to them"""
    def __init__(self, Nsamples=200):
        super().__init__()
        self._datas = dict([])
        self._Nsamples = Nsamples
        self._xaxis = None
        self._data_length = 0

    @property
    def size(self):
        return self._data_length

    @property
    def length(self):
        return self._Nsamples

    @length.setter
    def length(self, history_length: int):
        if history_length > 0:
            self._Nsamples = history_length

    def __len__(self):
        return self.length

    @dispatch(data_mod.DataWithAxes)
    def add_datas(self, data: data_mod.DataRaw):
        datas = {data.labels[ind]: data.data[ind] for ind in range(len(data))}
        self.add_datas(datas)

    @dispatch(list)
    def add_datas(self, datas: list):
        """
        Add datas to the history
        Parameters
        ----------
        datas: (list) list of floats or np.array(float)
        """
        datas = {f'data_{ind:02d}': datas[ind] for ind in range(len(datas))}
        self.add_datas(datas)

    @dispatch(dict)
    def add_datas(self, datas: dict):
        """
        Add datas to the history on the form of a dict of key/data pairs (data is a numpy 0D array)
        Parameters
        ----------
        datas: (dict) dictionaary of floats or np.array(float)
        """
        if datas.keys() != self._datas.keys():
            self.clear_data()

        self._data_length += 1

        if self._data_length > self._Nsamples:
            self._xaxis += 1
        else:
            self._xaxis = np.linspace(0, self._data_length, self._data_length, endpoint=False)

        for data_key, data in datas.items():
            if not isinstance(data, np.ndarray):
                data = np.array([data])

            if self._data_length == 1:
                self._datas[data_key] = data
            else:
                self._datas[data_key] = np.concatenate((self._datas[data_key], data))

            if self._data_length > self._Nsamples:
                self._datas[data_key] = self._datas[data_key][1:]

    @property
    def datas(self):
        return self._datas

    @property
    def xaxis(self):
        return self._xaxis

    def clear_data(self):
        self._datas = dict([])
        self._data_length = 0
        self._xaxis = np.array([])


class View_cust(pg.ViewBox):
    """Custom ViewBox used to enable other properties compared to parent class: pg.ViewBox

    """
    sig_double_clicked = QtCore.Signal(float, float)

    def __init__(self, parent=None, border=None, lockAspect=False, enableMouse=True, invertY=False,
                 enableMenu=True, name=None, invertX=False):
        super().__init__(parent, border, lockAspect, enableMouse, invertY, enableMenu, name, invertX)

    def mouseClickEvent(self, ev):
        if ev.button() == QtCore.Qt.RightButton and self.menuEnabled():
            ev.accept()
            self.raiseContextMenu(ev)
        if ev.double():
            pos = self.mapToView(ev.pos())
            self.sig_double_clicked.emit(pos.x(), pos.y())



# -*- coding: utf-8 -*-
"""
Created the 07/11/2022

@author: Sebastien Weber
"""
from typing import Union, List, TYPE_CHECKING
import numpy as np

if TYPE_CHECKING:
    from pymodaq.utils.data import DataWithAxes, Axis


class SpecialSlicers(object):
    """make it elegant to apply a slice to navigation or signal dimensions"""
    def __init__(self, obj: Union['DataWithAxes', 'Axis'], is_navigation):
        self.is_navigation = is_navigation
        self.obj = obj

    def __getitem__(self, slices):
        return self.obj._slicer(slices, self.is_navigation)


class SpecialSlicersData(SpecialSlicers):

    def __setitem__(self, slices, data: Union[np.ndarray, 'DataWithAxes', 'Axis']):
        """x.__setitem__(slices, data) <==> x[slices]=data
        """
        slices = self.obj._compute_slices(slices, self.is_navigation)

        if hasattr(self.obj, 'units'):
            if isinstance(data, np.ndarray):
                data_to_replace = data
            else:
                data_to_replace = data.get_data()
            if hasattr(self.obj, 'units') and self.obj.data is None:
                self.obj.create_linear_data(len(self.obj))
            self.obj.data[slices] = data_to_replace
        else:
            for ind in range(len(self.obj)):
                if isinstance(data, np.ndarray):
                    data_to_replace = data
                else:  # means it's a DataWithAxes
                    data_to_replace = data[ind]
                self.obj[ind][slices] = data_to_replace

    def __len__(self):
        return self.obj.axes_manager.sig_shape[0]


if __name__ == '__main__':
    from pymodaq.utils.data import DataWithAxes, DataRaw, Axis
    shape = (4, 10, 5, 7)
    dat = np.arange(np.prod(shape))
    dat = dat.reshape(shape)
    data = DataRaw('mydata', data=[dat], nav_indexes=(0, 1, 2),
                   axes=[Axis(f'axis_{ind:02d}',
                              data=np.linspace(0, shape[ind]-1, shape[ind]),
                              index=ind) for ind in range(len(shape))])
    subdata_sig = data.inav[2, 2, 3].data
    subdata_nav = data.isig[3].data
    data.isig[:]
    pass



from pathlib import Path
from typing import List, Union, Dict

from qtpy import QtWidgets, QtCore
from pymodaq.utils.managers.action_manager import ActionManager
from pymodaq.utils.parameter import Parameter, ParameterTree, ioxml, utils
from pymodaq.utils.gui_utils.file_io import select_file
from pymodaq.utils.config import get_set_config_dir

from pymodaq.utils.logger import set_logger, get_module_name
logger = set_logger(get_module_name(__file__))

class ParameterTreeWidget(ActionManager):

    def __init__(self, action_list: tuple = ('save', 'update', 'load')):
        super().__init__()

        self.widget = QtWidgets.QWidget()
        self.widget.setLayout(QtWidgets.QVBoxLayout())


        toolbar = QtWidgets.QToolBar()
        self.set_toolbar(toolbar)
        self.tree: ParameterTree = ParameterTree()

        self.widget.header = self.tree.header  # for back-compatibility, widget behave a bit like a ParameterTree
        self.widget.listAllItems = self.tree.listAllItems  # for back-compatibility

        self.tree.setMinimumWidth(150)
        self.tree.setMinimumHeight(300)
        
        # Making the buttons
        self.setup_actions(action_list) 
        # Making the splitter
        self.splitter = QtWidgets.QSplitter(QtCore.Qt.Vertical)
        # Adding the toolbar + the parameter tree
        self.splitter.addWidget(toolbar)
        self.splitter.addWidget(self.tree)
        # Hiding toolbar
        self.splitter.setSizes([0, 300])
        # Adding splitter to layout
        self.widget.layout().addWidget(self.splitter)
        self.widget.layout().setContentsMargins(0, 0, 0, 0)            

        


    def setup_actions(self, action_list: tuple = ('save', 'update', 'load')):
        """
        See Also
        --------
        ActionManager.add_action
        """
        # Saving action
        self.add_action('save_settings', 'Save Settings', 'saveTree',
                        "Save current settings in an xml file", 
                        visible = 'save' in action_list)
        # Update action
        self.add_action('update_settings', 'Update Settings', 'updateTree',
                        "Update the settings from an xml file, the settings structure loaded must be identical to the current one",
                        visible = 'update' in action_list)                
        # Load action
        self.add_action('load_settings', 'Load Settings', 'openTree',
                        "Load current settings from an xml file, the current settings structure is erased and is replaced by the new one",
                         visible = 'load' in action_list)


class ParameterManager:
    """Class dealing with Parameter and ParameterTree

    Attributes
    ----------
    params: list of dicts
        Defining the Parameter tree like structure
    settings_name: str
        The particular name to give to the object parent Parameter (self.settings)
    settings: Parameter
        The higher level (parent) Parameter
    settings_tree: QWidget
        widget Holding a ParameterTree and a toolbar for interacting with the tree
    tree: ParameterTree
        the underlying ParameterTree
    """
    settings_name = 'custom_settings'
    params = []

    def __init__(self, settings_name: str = None, action_list: tuple = ('save', 'update', 'load')):
        if settings_name is None:
            settings_name = self.settings_name
        # create a settings tree to be shown eventually in a dock
        # object containing the settings defined in the preamble
        # create a settings tree to be shown eventually in a dock
        self._settings_tree = ParameterTreeWidget(action_list)
        
        self._settings_tree.get_action(f'save_settings').connect_to(self.save_settings_slot)
        self._settings_tree.get_action(f'update_settings').connect_to(self.update_settings_slot)
        self._settings_tree.get_action(f'load_settings').connect_to(self.load_settings_slot)
                                                                        
        self.settings: Parameter = Parameter.create(name=settings_name, type='group', children=self.params)  # create a Parameter
        # object containing the settings defined in the preamble

    @property
    def settings_tree(self):
        return self._settings_tree.widget

    @property
    def tree(self):
        return self._settings_tree.tree

    @property
    def settings(self):
        return self._settings

    @settings.setter
    def settings(self, settings: Union[Parameter, List[Dict[str, str]], Path]):
        settings = self.create_parameter(settings)
        self._settings = settings
        self.tree.setParameters(self._settings, showTop=False)  # load the tree with this parameter object
        self._settings.sigTreeStateChanged.connect(self.parameter_tree_changed)

    @staticmethod
    def create_parameter(settings: Union[Parameter, List[Dict[str, str]], Path]) -> Parameter:

        if isinstance(settings, List):
            _settings = Parameter.create(title='Settings', name='settings', type='group', children=settings)
        elif isinstance(settings, Path) or isinstance(settings, str):
            settings = Path(settings)
            _settings = Parameter.create(title='Settings', name='settings',
                                        type='group', children=ioxml.XML_file_to_parameter(str(settings)))
        elif isinstance(settings, Parameter):
            _settings = Parameter.create(title='Settings', name=settings.name(), type='group')
            _settings.restoreState(settings.saveState())
        else:
            raise TypeError(f'Cannot create Parameter object from {settings}')
        return _settings

    def parameter_tree_changed(self, param, changes):
        for param, change, data in changes:
            path = self._settings.childPath(param)
            if change == 'childAdded':
                self.child_added(param, data)

            elif change == 'value':
                self.value_changed(param)

            elif change == 'parent':
                self.param_deleted(param)

    def value_changed(self, param):
        """Non-mandatory method  to be subclassed for actions to perform (methods to call) when one of the param's
        value in self._settings is changed

        Parameters
        ----------
        param: Parameter
            the parameter whose value just changed

        Examples
        --------
        >>> if param.name() == 'do_something':
        >>>     if param.value():
        >>>         print('Do something')
        >>>         self.settings.child('main_settings', 'something_done').setValue(False)
        """
        ...

    def child_added(self, param, data):
        """Non-mandatory method to be subclassed for actions to perform when a param has been added in self.settings

        Parameters
        ----------
        param: Parameter
            the parameter where child will be added
        data: Parameter
            the child parameter
        """
        pass

    def param_deleted(self, param):
        """Non-mandatory method to be subclassed for actions to perform when one of the param in self.settings has been deleted

        Parameters
        ----------
        param: Parameter
            the parameter that has been deleted
        """
        pass

    def save_settings_slot(self, file_path: Path = None):
        """ Method to save the current settings using a xml file extension.

        The starting directory is the user config folder with a subfolder called settings folder

        Parameters
        ----------
        file_path: Path
            Path like object pointing to a xml file encoding a Parameter object
            If None, opens a file explorer window to save manually a file
        """
        if file_path is None or file_path is False:
            file_path = select_file(get_set_config_dir('settings', user=True), save=True, ext='xml', filter='*.xml',
                                    force_save_extension=True)
        else:
            file_path = Path(file_path)
            if '.xml' != file_path.suffix:
                return
        if file_path:
            ioxml.parameter_to_xml_file(self.settings, file_path.resolve())
            logger.info(f'The settings have been successfully saved at {file_path}')

    def _get_settings_from_file(self):
        return select_file(get_set_config_dir('settings', user=True), save=False, ext='xml', filter='*.xml',
                           force_save_extension=True)

    def load_settings_slot(self, file_path: Path = None):
        """ Method to load settings into the parameter using a xml file extension.

        The starting directory is the user config folder with a subfolder called settings folder

        Parameters
        ----------
        file_path: Path
            Path like object pointing to a xml file encoding a Parameter object
            If None, opens a file explorer window to pick manually a file
        """
        if file_path is None or file_path is False:
            file_path = self._get_settings_from_file()
        if file_path:
            self.settings = file_path.resolve()
            logger.info(f'The settings from {file_path} have been successfully loaded')            

    def update_settings_slot(self, file_path: Path = None):
        """ Method to update settings using a xml file extension.

        The file should define the same settings structure (names and children)

        The starting directory is the user config folder with a subfolder called settings folder

        Parameters
        ----------
        file_path: Path
            Path like object pointing to a xml file encoding a Parameter object
            If None, opens a file explorer window to pick manually a file
        """
        if file_path is None or file_path is False:
            file_path = self._get_settings_from_file()
        if file_path:
            _settings = self.create_parameter(file_path.resolve())
            # Checking if both parameters have the same structure
            sameStruct = utils.compareStructureParameter(self.settings,_settings)
            if sameStruct:  # Update if true
                self.settings = _settings
                logger.info(f'The settings from {file_path} have been successfully applied')                            
            else:
                logger.info(f'The loaded settings from {file_path} do not match the current settings structure and cannot be applied.')



if __name__ == '__main__':

    class RealParameterManager(ParameterManager):
        params = {'title': 'Numbers:', 'name': 'numbers', 'type': 'group', 'children': [
            {'title': 'Standard float', 'name': 'afloat', 'type': 'float', 'value': 20., 'min': 1.,
             'tip': 'displays this text as a tooltip'},
            {'title': 'Linear Slide float', 'name': 'linearslidefloat', 'type': 'slide', 'value': 50, 'default': 50,
             'min': 0,
             'max': 123, 'subtype': 'linear'},
            {'title': 'Log Slide float', 'name': 'logslidefloat', 'type': 'slide', 'value': 50, 'default': 50,
             'min': 1e-5,
             'max': 1e5, 'subtype': 'log'},
        ]},


    import sys
    from qtpy import QtWidgets
    app = QtWidgets.QApplication(sys.argv)
    param_manager = RealParameterManager()
    param_manager.settings_tree.show()
    sys.exit(app.exec())




from enum import Enum
from pymodaq.utils.daq_utils import find_index
from pymodaq.utils.math_utils import my_moment
import numpy as np
from scipy.optimize import curve_fit
from typing import List, Union


class BaseEnum(Enum):
    """Enum to be used within pymodaq with some utility methods"""

    @classmethod
    def names(cls) -> List[str]:
        """Returns all the names of the enum"""
        return list(cls.__members__.keys())

    @classmethod
    def values(cls) -> List[str]:
        """Returns all the names of the enum"""
        return [cls[name].value for name in cls.names()]

    def __eq__(self, other: Union[str, Enum]):
        """testing for equality using the enum name"""
        if isinstance(other, str):
            if other == self.name:
                return True
        return super().__eq__(other)


def enum_checker(enum: BaseEnum, item: Union[BaseEnum, str]):
    """Check if the item parameter is a valid enum or at least one valid string name of the enum

    If a string, transforms it to a valid enum (case not important)

    Parameters
    ----------
    enum: BaseEnum class or one of its derivated class

    item: str or BaseEnum instance

    Returns
    -------
    BaseEnum class or one of its derivated class
    """

    if not isinstance(item, enum):
        if not isinstance(item, str):
            raise ValueError(f'{item} is an invalid {enum}. Should be a {enum} enum or '
                             f'a string in {enum.names()}')
        for ind, name in enumerate(enum.names()):
            if item.lower() == name.lower():
                item = enum[name]
                break
        else:
            raise ValueError(f'{item} is an invalid {enum}. Should be a {enum} enum or '
                             f'a string in {enum.names()}')
    return item






from abc import abstractproperty

from os import environ
import sys
import datetime
from pathlib import Path
from typing import Union, Dict, TypeVar, Any

import toml


try:
    USER = environ['USERNAME'] if sys.platform == 'win32' else environ['USER']
except:
    USER = 'unknown_user'

CONFIG_BASE_PATH = Path(environ['PROGRAMDATA']) if sys.platform == 'win32' else \
    Path('Library/Application Support') if sys.platform == 'darwin' else Path('/etc')


KeyType = TypeVar('KeyType')


def deep_update(mapping: Dict[KeyType, Any], *updating_mappings: Dict[KeyType, Any]) -> Dict[KeyType, Any]:
    """ Make sure a dictionary is updated using another dict in any nested level
    Taken from Pydantic v1
    """
    updated_mapping = mapping.copy()
    for updating_mapping in updating_mappings:
        for k, v in updating_mapping.items():
            if k in updated_mapping and isinstance(updated_mapping[k], dict) and isinstance(v, dict):
                updated_mapping[k] = deep_update(updated_mapping[k], v)
            else:
                updated_mapping[k] = v
    return updated_mapping


def replace_file_extension(filename: str, ext: str):
    """Replace the extension of a file by the specified one, without the dot"""
    file_name = Path(filename).stem  # remove eventual extensions
    if ext[0] == '.':
        ext = ext[1:]
    file_name += '.' + ext
    return file_name


def getitem_recursive(dic, *args, ndepth=0, create_if_missing=False):
    """Will scan recursively a dictionary in order to get the item defined by the iterable args

    Parameters
    ----------
    dic: dict
        the dictionary to scan
    args: an iterable of str
        keys of the dict
    ndepth: int
        by default (0) get the last element defined by args. 1 would mean it get the parent dict, 2 the parent of the
        parent...
    create_if_missing: bool
        if the entry is not present, create it assigning the 'none' default value (as a lower case string)
    Returns
    -------
    object or dict
    """
    args = list(args)
    while len(args) > ndepth:
        try:
            arg = args.pop(0)
            dic = dic[arg]
        except KeyError as e:
            if create_if_missing:
                if len(args) > 0:
                    dic[arg] = {}
                    dic = dic[arg]
                else:
                    dic[arg] = 'none'
                    dic = 'none'
            else:
                raise e
    return dic


def get_set_path(a_base_path: Path, dir_name: str) -> Path:
    path_to_get = a_base_path.joinpath(dir_name)
    if not path_to_get.is_dir():
        try:
            path_to_get.mkdir()
        except PermissionError as e:
            print(f"Cannot create local config folder at this location: {path_to_get}"
                  f", try using admin rights. "
                  f"Changing the not permitted path to a user one: {Path.home().joinpath(dir_name)}.")
            path_to_get = Path.home().joinpath(dir_name)
            if not path_to_get.is_dir():
                path_to_get.mkdir()
    return path_to_get


def get_set_local_dir(user=False) -> Path:
    """Defines, creates and returns a local folder where configuration files will be saved

    Depending on the os the configurations files will be stored in CONFIG_BASE_PATH, then
    each user will have another one created that could override the default and system-wide base folder

    Parameters
    ----------
    user: bool
        if False get the system-wide folder, otherwise the user folder

    Returns
    -------
    Path: the local path
    """
    if user:
        local_path = get_set_path(Path.home(), '.pymodaq')
    else:
        local_path = get_set_path(CONFIG_BASE_PATH, '.pymodaq')
    return local_path


def get_config_file(config_file_name: str, user=False):
    return get_set_local_dir(user).joinpath(replace_file_extension(config_file_name, 'toml'))


def get_set_config_dir(config_name='config', user=False):
    """Creates a folder in the local config directory to store specific configuration files

    Parameters
    ----------
    config_name: (str) name of the configuration folder
    user: bool
        if False get the system-wide folder, otherwise the user folder

    Returns
    -------
    Path

    See Also
    --------
    get_set_local_dir
    """
    return get_set_path(get_set_local_dir(user=user), config_name)


def get_set_log_path():
    """ creates and return the config folder path for log files
    """
    return get_set_config_dir('log')


def get_set_preset_path():
    """ creates and return the config folder path for managers files
    """
    return get_set_config_dir('preset_configs')


def get_set_batch_path():
    """ creates and return the config folder path for managers files
    """
    return get_set_config_dir('batch_configs')


def get_set_pid_path():
    """ creates and return the config folder path for PID files
    """
    return get_set_config_dir('pid_configs')


def get_set_layout_path():
    """ creates and return the config folder path for layout files
    """
    return get_set_config_dir('layout_configs')


def get_set_remote_path():
    """ creates and return the config folder path for remote (shortcuts or joystick) files
    """
    return get_set_config_dir('remote_configs')


def get_set_overshoot_path():
    """ creates and return the config folder path for overshoot files
    """
    return get_set_config_dir('overshoot_configs')


def get_set_roi_path():
    """ creates and return the config folder path for managers files
    """
    return get_set_config_dir('roi_configs')


def create_toml_from_dict(mydict: dict, dest_path: Path):
    """Create a Toml file at a given path from a dictionnary"""
    dest_path.write_text(toml.dumps(mydict))


def check_config(config_base: dict, config_local: dict):
        """Compare two configuration dictionaries. Adding missing keys

        Parameters
        ----------
        config_base: dict
            The base dictionaries with possible new keys
        config_local: dict
            a dict from a local config file potentially missing keys

        Returns
        -------
        bool: True if keys where missing else False
        """
        status = False
        for key in config_base:
            if key in config_local:
                if isinstance(config_base[key], dict):
                    status = status or check_config(config_base[key], config_local[key])
            else:
                config_local[key] = config_base[key]
                status = True
        return status


def copy_template_config(config_file_name: str = 'config', source_path: Union[Path, str] = None,
                         dest_path: Union[Path, str] = None):
    """Get a toml file path and copy it

    the destination is made of a given folder path (or the system-wide local path by default) and the config_file_name
    appended by the suffix '.toml'

    The source file (or pymodaq config template path by default) is read and dumped in this destination file

    Parameters
    ----------
    config_file_name: str
        the name of the destination config file
    source_path: Path or str
        the path of the toml source to be copied
    dest_path: Path or str
        the destination path of the copied config

    Returns
    -------
    Path: the path of the copied file
    """
    if dest_path is None:
        dest_path = get_set_local_dir()

    file_name = Path(config_file_name).stem  # remove eventual extensions
    file_name += '.toml'
    dest_path_with_filename = dest_path.joinpath(file_name)

    if source_path is None:
        config_template_dict = toml.load(Path(__file__).parent.parent.joinpath('resources/config_template.toml'))
    else:
        config_template_dict = toml.load(Path(source_path))

    create_toml_from_dict(config_template_dict, dest_path_with_filename)
    return dest_path_with_filename


def load_system_config_and_update_from_user(config_file_name: str):
    """load from a system-wide config file, update it from the user config file

    Parameters
    ----------
    config_file_name: str
        The config file to be loaded
    Returns
    -------
    dict: contains the toml system-wide file update with the user file
    """
    config_dict = dict([])
    toml_base_path = get_config_file(config_file_name, user=False)
    if toml_base_path.is_file():
        config_dict = toml.load(toml_base_path)
    toml_user_path = get_config_file(config_file_name, user=True)
    if toml_user_path.is_file():
        config_dict = deep_update(config_dict, toml.load(toml_user_path))
    return config_dict


class ConfigError(Exception):
    pass


class BaseConfig:
    """Base class to manage configuration files

    Should be subclassed with proper class attributes for each configuration file you need with pymodaq

    Attributes
    ----------
    config_name: str
        The name with which the configuration will be saved
    config_template_path: Path
        The Path of the template from which the config is constructed

    """
    config_template_path: Path = abstractproperty()
    config_name: str = abstractproperty()

    def __init__(self):
        self._config = self.load_config(self.config_name, self.config_template_path)

    def __repr__(self):
        return f'{self.config_name} configuration file'

    def __call__(self, *args):
        try:
            ret = getitem_recursive(self._config, *args)
        except KeyError as e:
            raise ConfigError(f'the path {args} does not exist in your configuration toml file, check '
                              f'your pymodaq_local folder')
        return ret

    def to_dict(self):
        return self._config

    def __getitem__(self, item):
        """for backcompatibility when it was a dictionnary"""
        if isinstance(item, tuple):
            return getitem_recursive(self._config, *item)
        else:
            return self._config[item]

    def __setitem__(self, key, value):
        if isinstance(key, tuple):
            dic = getitem_recursive(self._config, *key, ndepth=1, create_if_missing=False)
            dic[key[-1]] = value
        else:
            self._config[key] = value

    def load_config(self, config_file_name, template_path: Path):
        """Load a configuration file from both system-wide and user file

        check also if missing entries in the configuration file compared to the template"""
        toml_base_path = get_config_file(config_file_name, user=False)
        toml_user_path = get_config_file(config_file_name, user=True)
        if toml_base_path.is_file():
            config = toml.load(toml_base_path)
            config_template = toml.load(template_path)
            if check_config(config_template, config):  # check if all fields from template are there
                # (could have been  modified by some commits)
                create_toml_from_dict(config, toml_base_path)

        else:
            copy_template_config(config_file_name, template_path, toml_base_path.parent)

        if not toml_user_path.is_file():
            # create the author from environment variable
            config_dict = self.dict_to_add_to_user()
            if config_dict is not None:
                create_toml_from_dict(config_dict, toml_user_path)

        config_dict = load_system_config_and_update_from_user(config_file_name)
        return config_dict

    def dict_to_add_to_user(self):
        """To subclass"""
        return dict([])

    @property
    def config_path(self):
        """Get the user config path"""
        return get_config_file(self.config_name, user=True)

    @property
    def system_config_path(self):
        """Get the system_wide config path"""
        return get_config_file(self.config_name, user=False)

    def save(self):
        """Save the current Config object into the user toml file"""
        self.config_path.write_text(toml.dumps(self.to_dict()))


class Config(BaseConfig):
    """Main class to deal with configuration values for PyMoDAQ"""
    config_template_path = Path(__file__).parent.parent.joinpath('resources/config_template.toml')
    config_name = 'config_pymodaq'

    def dict_to_add_to_user(self):
        """To subclass"""
        return dict(user=dict(name=USER))



    
if __name__ == '__main__':

    config = Config()
    config('style', 'darkstyle')
    assert config('style', 'darkstyle') == config['style']['darkstyle']




import pickle

from pymodaq.utils import config as config_mod
from pymodaq.utils.gui_utils.file_io import select_file


def load_layout_state(dockarea, file=None):
    """
        Load and restore a dockarea layout state from the select_file obtained pathname file.
    """
    if file is None:
        file = select_file(start_path=config_mod.get_set_layout_path(), save=False, ext='dock')
    if file is not None:
        with open(str(file), 'rb') as f:
            dockstate = pickle.load(f)
            dockarea.restoreState(dockstate)
    file = file.name
    return file


def save_layout_state(dockarea, file=None):
    """
        Save the current layout state in the select_file obtained pathname file.
        Once done dump the pickle.
    """

    dockstate = dockarea.saveState()
    if 'float' in dockstate:
        dockstate['float'] = []
    if file is None:
        file = select_file(start_path=config_mod.get_set_layout_path(), save=True, ext='dock')
    if file is not None:
        with open(str(file), 'wb') as f:
            pickle.dump(dockstate, f, pickle.HIGHEST_PROTOCOL)